import os
import torch
import torch.nn.functional as F
from timm.models.layers import DropPath, to_2tuple
from torch import nn

import torch.utils.checkpoint as checkpoint
from functools import partial
from einops import rearrange

# try:
#     from .flash_attention_class import FlashAttention
# except:
#     from flash_attn import FlashAttention
# from flash_attn.modules.mlp import FusedMLP
# from flash_attn.ops.rms_norm import DropoutAddRMSNorm


MODEL_PATH = 'your_model_path/internvl'
_MODELS = {
    # see InternVL
    "internvl_c_13b_224px": os.path.join(MODEL_PATH, "internvl_c_13b_224px.pth"),
}


class CrossAttention(nn.Module):
    def __init__(
            self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0.,
            proj_drop=0., attn_head_dim=None, out_dim=None):
        super().__init__()
        if out_dim is None:
            out_dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        if attn_head_dim is not None:
            head_dim = attn_head_dim
        all_head_dim = head_dim * self.num_heads
        self.scale = qk_scale or head_dim ** -0.5
        assert all_head_dim == dim
        
        self.q = nn.Linear(dim, all_head_dim, bias=False)
        self.k = nn.Linear(dim, all_head_dim, bias=False)
        self.v = nn.Linear(dim, all_head_dim, bias=False)
        
        if qkv_bias:
            self.q_bias = nn.Parameter(torch.zeros(all_head_dim))
            self.k_bias = nn.Parameter(torch.zeros(all_head_dim))
            self.v_bias = nn.Parameter(torch.zeros(all_head_dim))
        else:
            self.q_bias = None
            self.k_bias = None
            self.v_bias = None
        
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(all_head_dim, out_dim)
        self.proj_drop = nn.Dropout(proj_drop)
    
    def forward(self, x, k=None, v=None, return_attn=False):
        B, N, C = x.shape
        N_k = k.shape[1]
        N_v = v.shape[1]
        
        q_bias, k_bias, v_bias = None, None, None
        if self.q_bias is not None:
            q_bias = self.q_bias
            k_bias = self.k_bias
            v_bias = self.v_bias
        
        q = F.linear(input=x, weight=self.q.weight, bias=q_bias)
        q = q.reshape(B, N, 1, self.num_heads, -1).permute(2, 0, 3, 1, 4).squeeze(0)  # (B, N_head, N_q, dim)
        
        k = F.linear(input=k, weight=self.k.weight, bias=k_bias)
        k = k.reshape(B, N_k, 1, self.num_heads, -1).permute(2, 0, 3, 1, 4).squeeze(0)
        
        v = F.linear(input=v, weight=self.v.weight, bias=v_bias)
        v = v.reshape(B, N_v, 1, self.num_heads, -1).permute(2, 0, 3, 1, 4).squeeze(0)
        
        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))  # (B, N_head, N_q, N_k)
        
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        
        x = (attn @ v).transpose(1, 2).reshape(B, N, -1)
        x = self.proj(x)
        x = self.proj_drop(x)
        
        if return_attn:
            return x, attn.mean(1) # (B, n_head, n_q, C) => (B, n_q, C)
        else:
            return x, None


class AttentiveBlock(nn.Module):
    
    def __init__(self, dim, num_heads, qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., norm_layer=nn.LayerNorm, attn_head_dim=None, out_dim=None):
        super().__init__()
        
        self.norm1_q = norm_layer(dim)
        self.norm1_k = norm_layer(dim)
        self.norm1_v = norm_layer(dim)
        self.cross_attn = CrossAttention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop,
            proj_drop=drop, attn_head_dim=attn_head_dim, out_dim=out_dim)
        
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
    
    def forward(self, x_q, x_kv, pos_q, pos_k, bool_masked_pos, rel_pos_bias=None, return_attn=False):
        x_q = self.norm1_q(x_q + pos_q)
        x_k = self.norm1_k(x_kv + pos_k)
        x_v = self.norm1_v(x_kv)
        x, attn = self.cross_attn(x_q, k=x_k, v=x_v, return_attn=return_attn)
        return x, attn


class AttentionPoolingBlock(AttentiveBlock):
    
    def forward(self, x, return_attn=False):
        x_q = x.mean(1, keepdim=True)
        x_kv, pos_q, pos_k = x, 0, 0
        x, attn = super().forward(x_q, x_kv, pos_q, pos_k, bool_masked_pos=None, rel_pos_bias=None, return_attn=return_attn)
        x = x.squeeze(1)
        return x, attn


class RMSNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.variance_epsilon = eps
    
    def forward(self, hidden_states):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states.to(input_dtype)


class LayerScale(nn.Module):
    def __init__(self, dim, init_values=1e-5, inplace=False, force_fp32=False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))
        self.force_fp32 = force_fp32
    
    # @torch.cuda.amp.autocast(enabled=False)
    def forward(self, x):
        if self.force_fp32:
            output_type = x.dtype
            out = x.float().mul_(self.gamma.float()) if self.inplace else x.float() * self.gamma.float()
            return out.to(dtype=output_type)
        else:
            out = x.mul_(self.gamma) if self.inplace else x * self.gamma
            return out


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0., proj_drop=0., use_flash_attn=False,
                 causal=False, norm_layer=nn.LayerNorm, qk_normalization=False, use_fused_rmsnorm=False):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5
        
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        
        self.use_flash_attn = use_flash_attn
        # if use_flash_attn:
        #     self.causal = causal
        #     self.inner_attn = FlashAttention(attention_dropout=attn_drop)
        
        self.qk_normalization = qk_normalization
        self.q_norm = norm_layer(dim) if qk_normalization else nn.Identity()
        self.k_norm = norm_layer(dim) if qk_normalization else nn.Identity()
        self.use_fused_rmsnorm = use_fused_rmsnorm
    
    def _naive_attn(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)  # make torchscript happy (cannot use tensor as tuple)
        
        if self.qk_normalization:
            B_, H_, N_, D_ = q.shape
            q = self.q_norm(q.transpose(1, 2).flatten(-2, -1)).view(B_, N_, H_, D_).transpose(1, 2)
            k = self.k_norm(k.transpose(1, 2).flatten(-2, -1)).view(B_, N_, H_, D_).transpose(1, 2)
        
        attn = ((q * self.scale) @ k.transpose(-2, -1))
        # attn = attn - attn.max(-1)[0].unsqueeze(-1)  # in case of overflow for fp16
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        
        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x
    
    def _flash_attn(self, x, key_padding_mask=None, need_weights=False):
        
        qkv = self.qkv(x)
        qkv = rearrange(qkv, "b s (three h d) -> b s three h d", three=3, h=self.num_heads)
        
        if self.qk_normalization:
            q, k, v = qkv.unbind(2)
            if self.use_fused_rmsnorm:
                q = self.q_norm(q.flatten(-2, -1))[0].view(q.shape)
                k = self.k_norm(k.flatten(-2, -1))[0].view(k.shape)
            else:
                q = self.q_norm(q.flatten(-2, -1)).view(q.shape)
                k = self.k_norm(k.flatten(-2, -1)).view(k.shape)
            qkv = torch.stack([q, k, v], dim=2)
        
        context, _ = self.inner_attn(
            qkv, key_padding_mask=key_padding_mask, need_weights=need_weights, causal=self.causal
        )
        outs = self.proj(rearrange(context, "b s h d -> b s (h d)"))
        outs = self.proj_drop(outs)
        return outs
    
    def forward(self, x):
        x = self._naive_attn(x) if not self.use_flash_attn else self._flash_attn(x)
        return x


class Mlp(nn.Module):
    """ MLP as used in Vision Transformer, MLP-Mixer and related networks
    """
    
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU,
                 bias=True, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = to_2tuple(bias)
        drop_probs = to_2tuple(drop)
        
        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])
    
    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)
        return x


class Block(nn.Module):
    
    def __init__(
            self, dim, num_heads, mlp_ratio=4., qkv_bias=False, drop=0., attn_drop=0., init_values=None,
            drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, use_flash_attn=False, use_fused_mlp=False,
            fused_mlp_heuristic=1, with_cp=False, qk_normalization=False, layerscale_no_force_fp32=False,
            use_fused_rmsnorm=False):
        super().__init__()
        
        self.norm1 = norm_layer(dim)
        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop,
                              use_flash_attn=use_flash_attn, causal=False, norm_layer=norm_layer,
                              qk_normalization=qk_normalization,
                              use_fused_rmsnorm=use_fused_rmsnorm)
        self.ls1 = LayerScale(dim, init_values=init_values,
                              force_fp32=(not layerscale_no_force_fp32)) if init_values else nn.Identity()
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path1 = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        # if use_fused_mlp:
        #     self.mlp = FusedMLP(in_features=dim, hidden_features=mlp_hidden_dim, heuristic=fused_mlp_heuristic)
        # else:
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)
        self.ls2 = LayerScale(dim, init_values=init_values,
                              force_fp32=(not layerscale_no_force_fp32)) if init_values else nn.Identity()
        self.drop_path2 = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        
        self.with_cp = with_cp
        self.use_fused_rmsnorm = use_fused_rmsnorm
    
    def forward(self, x, residual=None):
        
        def _inner_forward(x, residual=None):
            if self.use_fused_rmsnorm:
                x, residual = self.norm1(x, residual)
                x = self.drop_path1(self.ls1(self.attn(x)))
                x, residual = self.norm2(x, residual)
                x = self.drop_path2(self.ls2(self.mlp(x)))
                return x, residual
            else:
                assert residual is None
                x = x + self.drop_path1(self.ls1(self.attn(self.norm1(x))))
                x = x + self.drop_path2(self.ls2(self.mlp(self.norm2(x))))
                return x
        
        if self.with_cp:
            return checkpoint.checkpoint(_inner_forward, x, residual)
        else:
            return _inner_forward(x, residual=residual)


class PatchEmbed(nn.Module):
    """ 3D Image to Patch Embedding
    """
    
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768, norm_layer=None, flatten=True):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.grid_size = (img_size[0] // patch_size[0], img_size[1] // patch_size[1])
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        self.flatten = flatten
        
        self.proj = nn.Conv3d(
            in_chans, embed_dim, 
            kernel_size=(1, patch_size[0], patch_size[1]), 
            stride=(1, patch_size[0], patch_size[1]), 
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()
    
    def forward(self, x):
        x = self.proj(x)
        if self.flatten:
            x = x.flatten(3).permute(0, 2, 3, 1)  # (N, C, T, H, W) => (N, T, H * W, C)
        x = self.norm(x)
        return x


class InternVL_CLIP(nn.Module):
    def __init__(
            self,
            in_chans: int = 3,
            patch_size: int = 14,
            img_size: int = 224,
            qkv_bias: bool = False,
            drop_path_rate: float = 0.2,
            embed_dim: int = 3200,
            num_heads: int = 25,
            mlp_ratio: int = 4,
            init_values: float = 0.1,
            qk_normalization: bool = True,
            depth: int = 48,
            use_flash_attn: bool = True,
            use_fused_rmsnorm: bool = True,
            use_fused_mlp: bool = True,
            fused_mlp_heuristic: int = 1,
            with_cp: bool = False,
            attn_pool_num_heads: int = 16,
            clip_embed_dim: int = 768,
            layerscale_no_force_fp32: bool = True,
            # for unmasked teacher
            clip_norm_type: str = 'l2',
            return_attn: bool = True,
            clip_return_layer: int = 1,
            clip_return_interval: int = 1,
            ):
        super().__init__()
        
        assert use_flash_attn == use_fused_rmsnorm == use_fused_mlp, print(
            'use_flash_attn, use_fused_rmsnorm and use_fused_mlp should be consistent')
        
        self.use_flash_attn = use_flash_attn
        self.embed_dim = embed_dim

        self.clip_norm_type = clip_norm_type
        self.return_attn = return_attn
        self.return_index = []
        for i in range(clip_return_layer):
            self.return_index.append(depth - int(i * clip_return_interval) - 1)
        print(f'Normalization Type: {clip_norm_type}')
        print(f'Return Attention: {return_attn}')
        print(f'Teacher Return Interval: {self.return_index}')
        
        """ only use image encoder of InternVL """
        # if use_fused_rmsnorm:
        #     norm_layer_for_blocks = partial(DropoutAddRMSNorm, eps=1e-6, prenorm=True)
        # else:
        norm_layer_for_blocks = partial(RMSNorm, eps=1e-6)
        self.norm_layer_for_blocks = norm_layer_for_blocks
        self.patch_embed = PatchEmbed(img_size, patch_size, in_chans, embed_dim)
        num_patches = self.patch_embed.num_patches
        self.num_patches = num_patches
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim), requires_grad=False)
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]
        
        self.blocks = nn.ModuleList([
            Block(embed_dim, num_heads, mlp_ratio, qkv_bias=qkv_bias,
                  norm_layer=norm_layer_for_blocks,
                  drop_path=dpr[i], init_values=init_values, attn_drop=0.,
                  use_flash_attn=use_flash_attn, use_fused_mlp=use_fused_mlp,
                  fused_mlp_heuristic=fused_mlp_heuristic,
                  with_cp=with_cp,
                  qk_normalization=qk_normalization,
                  layerscale_no_force_fp32=layerscale_no_force_fp32,
                  use_fused_rmsnorm=use_fused_rmsnorm)
            for i in range(depth)])
        self.clip_projector = AttentionPoolingBlock(
            dim=embed_dim, num_heads=attn_pool_num_heads, qkv_bias=True, qk_scale=None,
            drop=0., attn_drop=0., norm_layer=partial(nn.LayerNorm, eps=1e-5), out_dim=clip_embed_dim)
    
    @property
    def dtype(self):
        return self.patch_embed.proj.weight.dtype
    
    def forward(self, image):
        x = self.patch_embed(image.type(self.dtype))
        B, T, HW, C = x.size()
        x = x.reshape(B * T, HW, C)
        
        cls_tokens = self.cls_token.expand(B * T, -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)
        x = x + self.pos_embed

        residual = None
        z = []
        for idx, blk in enumerate(self.blocks):
            if isinstance(x, tuple) and len(x) == 2:
                x, residual = x
            x = blk(x, residual=residual)
            # return intermediate features
            if idx in self.return_index:
                if isinstance(x, tuple) and len(x) == 2:
                    tmp_x, tmp_residual = x
                    if residual is not None:
                        z.append(tmp_x + tmp_residual)
                else:
                    z.append(x)

        if isinstance(x, tuple) and len(x) == 2:
            x, residual = x
            if residual is not None:
                x = x + residual

        x, attn = self.clip_projector(x, return_attn=self.return_attn)

        if self.clip_norm_type == 'l2':
            # normalization of intermediate features
            z = torch.stack(z) # (K, BT, HW+1, C)
            K = z.shape[0]
            cls_tokens, z = z[:, :, :1, :], z[:, :, 1:, :]
            cls_tokens = cls_tokens.view(K, B, T, 1, C).mean(2) # (K, BT, 1, C) => (K, B, 1, C)
            z = z.reshape(K, B, T * HW, C)
            z = torch.cat((cls_tokens, z), dim=2) # (K, B, HWT+1, C)
            z = z / z.norm(dim=-1, keepdim=True)
            # normalization of final features
            x = x.view(B, T, -1).mean(1) # (BT, C) => (B, C)
            x = x / x.norm(dim=-1, keepdim=True)
        elif self.clip_norm_type == 'none':
            pass
        else:
            raise NotImplementedError

        if self.return_attn:
            return z, x, attn[:, 0, 1:] # (B * T, HW)
        else:
            return z, x


def inflate_weight(weight_2d, time_dim, center=True):
    print(f'Init center: {center}')
    if center:
        weight_3d = torch.zeros(*weight_2d.shape)
        weight_3d = weight_3d.unsqueeze(2).repeat(1, 1, time_dim, 1, 1)
        middle_idx = time_dim // 2
        weight_3d[:, :, middle_idx, :, :] = weight_2d
    else:
        weight_3d = weight_2d.unsqueeze(2).repeat(1, 1, time_dim, 1, 1)
        weight_3d = weight_3d / time_dim
    return weight_3d


def process_checkpoint(ckpt, model):
    new_ckpt = {}
    state_dict_3d = model.state_dict()
    for k, v in ckpt['module'].items():
        new_k = k
        if 'patch_embed' in new_k and new_k in state_dict_3d.keys() and v.shape != state_dict_3d[new_k].shape:
            print(new_k)
            print(f'Inflate: {k}, {v.shape} => {state_dict_3d[new_k].shape}')
            time_dim = state_dict_3d[new_k].shape[2]
            v = inflate_weight(v, time_dim)
        new_ckpt[new_k] = v

    # interpolate position embedding
    pos_embed_checkpoint = new_ckpt['pos_embed']
    embedding_size = pos_embed_checkpoint.shape[-1]
    num_patches = model.num_patches
    orig_size = int((pos_embed_checkpoint.shape[-2] - 1) ** 0.5)
    new_size = int(num_patches ** 0.5)
    if orig_size != new_size:
        print(f'pos_embed from {orig_size} to {new_size}')
        extra_tokens = pos_embed_checkpoint[:, :1]
        pos_tokens = pos_embed_checkpoint[:, 1:]
        pos_tokens = pos_tokens.reshape(-1, orig_size, orig_size, embedding_size).permute(0, 3, 1, 2)
        pos_tokens = torch.nn.functional.interpolate(
            pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False)
        pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(0, 2).unsqueeze(0)
        new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
        new_ckpt['pos_embed'] = new_pos_embed

    return new_ckpt


def internvl_clip_6b(
        img_size, 
        clip_norm_type='l2',
        return_attn=True,
        clip_return_layer=1,
        clip_return_interval=1
    ):
    model = InternVL_CLIP(
        img_size=img_size, 
        layerscale_no_force_fp32=False,
        clip_norm_type=clip_norm_type,
        return_attn=return_attn,
        clip_return_layer=clip_return_layer,
        clip_return_interval=clip_return_interval,
    )

    ckpt = torch.load(_MODELS["internvl_c_13b_224px"], map_location='cpu')
    new_ckpt = process_checkpoint(ckpt, model)
    message = model.load_state_dict(new_ckpt, strict=False)
    print(message)
    return model.eval()


if __name__ == '__main__':
    import time
    from fvcore.nn import FlopCountAnalysis
    from fvcore.nn import flop_count_table
    import numpy as np

    seed = 4217
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    num_frames = 8
    img_size = 224
    video = torch.rand(1, 3, num_frames, img_size, img_size).cuda().half()

    model = internvl_clip_6b(img_size).cuda().half()
    # flops = FlopCountAnalysis(model, video)
    model(video)




"""Tests for deduplication module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_deduplication_basics():
    from ayase.modules.dedup import DeduplicationModule
    _test_module_basics(DeduplicationModule, "deduplication")

def test_deduplication_image(image_sample):
    from ayase.modules.dedup import DeduplicationModule
    image_sample.quality_metrics = QualityMetrics()
    m = DeduplicationModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_deduplication_video(video_sample):
    from ayase.modules.dedup import DeduplicationModule
    video_sample.quality_metrics = QualityMetrics()
    m = DeduplicationModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Profile loader and module builder for app-specific Ayase pipelines.

This module intentionally does not ship hardcoded metric presets.
Applications define their own profiles (JSON/TOML or dict) and build
module instances from them.
"""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

from pydantic import BaseModel, Field

from .config import AyaseConfig
from .pipeline import ModuleRegistry, PipelineModule


class PipelineProfile(BaseModel):
    """External profile schema for assembling Ayase pipelines."""

    name: Optional[str] = None
    modules: List[str] = Field(default_factory=list)
    module_config: Dict[str, Dict[str, Any]] = Field(default_factory=dict)


def load_profile(profile: Union[Path, str, Dict[str, Any], PipelineProfile]) -> PipelineProfile:
    """Load a profile from an object, dict, JSON, or TOML file."""

    if isinstance(profile, PipelineProfile):
        return profile

    if isinstance(profile, dict):
        return cast(PipelineProfile, PipelineProfile.model_validate(profile))

    if isinstance(profile, str):
        profile = Path(profile)

    if not isinstance(profile, Path):
        raise TypeError("profile must be Path, str, dict, or PipelineProfile")

    if not profile.exists():
        raise FileNotFoundError(f"Profile file not found: {profile}")

    suffix = profile.suffix.lower()
    raw = profile.read_text(encoding="utf-8")

    if suffix == ".json":
        data = cast(Dict[str, Any], json.loads(raw))
    elif suffix in (".toml", ".tml"):
        try:
            import tomllib  # py3.11+
        except ModuleNotFoundError:
            import tomli as tomllib  # py3.9/3.10 fallback
        data = cast(Dict[str, Any], tomllib.loads(raw))
    else:
        raise ValueError(f"Unsupported profile format: {suffix}. Use .json or .toml")

    return cast(PipelineProfile, PipelineProfile.model_validate(data))


def instantiate_profile_modules(
    profile: Union[Path, str, Dict[str, Any], PipelineProfile],
    config: Optional[AyaseConfig] = None,
) -> List[PipelineModule]:
    """Instantiate modules for an app-defined profile."""

    loaded = load_profile(profile)
    cfg = config or AyaseConfig.load()

    ModuleRegistry.discover_modules(plugin_paths=cfg.pipeline.plugin_folders)

    if not loaded.modules:
        return []

    modules: List[PipelineModule] = []
    for module_name in loaded.modules:
        module_cls = ModuleRegistry.get_module(module_name)
        if module_cls is None:
            raise ValueError(f"Unknown module in profile: {module_name}")

        params: Dict[str, Any] = {
            "models_dir": str(cfg.general.models_dir),
        }

        per_module = loaded.module_config.get(module_name, {})
        if per_module:
            params.update(per_module)

        modules.append(module_cls(config=params))

    return modules



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

import copy
from typing import Optional, List
import math

import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torch.nn.init import xavier_uniform_, constant_, uniform_, normal_

from util.misc import inverse_sigmoid
from models.ops.modules import MSDeformAttn

from torchvision.ops.boxes import batched_nms
from util.box_ops import box_cxcywh_to_xyxy

class DeformableTransformer(nn.Module):
    def __init__(self, d_model=256, nhead=8,
                 num_encoder_layers=6, num_decoder_layers=6, dim_feedforward=1024, dropout=0.1,
                 activation="relu", return_intermediate_dec=False,
                 num_feature_levels=4, dec_n_points=4,  enc_n_points=4,
                 two_stage=False, two_stage_num_proposals=300,
                 assign_first_stage=False):
        super().__init__()

        self.d_model = d_model
        self.nhead = nhead
        self.two_stage = two_stage
        self.two_stage_num_proposals = two_stage_num_proposals
        self.assign_first_stage = assign_first_stage

        encoder_layer = DeformableTransformerEncoderLayer(d_model, dim_feedforward,
                                                          dropout, activation,
                                                          num_feature_levels, nhead, enc_n_points)
        self.encoder = DeformableTransformerEncoder(encoder_layer, num_encoder_layers)

        decoder_layer = DeformableTransformerDecoderLayer(d_model, dim_feedforward,
                                                          dropout, activation,
                                                          num_feature_levels, nhead, dec_n_points)
        self.decoder = DeformableTransformerDecoder(decoder_layer, num_decoder_layers, return_intermediate_dec)

        self.level_embed = nn.Parameter(torch.Tensor(num_feature_levels, d_model))

        if two_stage:
            self.enc_output = nn.Linear(d_model, d_model)
            self.enc_output_norm = nn.LayerNorm(d_model)
            self.pos_trans = nn.Linear(d_model * 2, d_model * 2)
            self.pos_trans_norm = nn.LayerNorm(d_model * 2)
            self.pix_trans = nn.Linear(d_model, d_model)
            self.pix_trans_norm = nn.LayerNorm(d_model)
        else:
            self.reference_points = nn.Linear(d_model, 2)

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)
        for m in self.modules():
            if isinstance(m, MSDeformAttn):
                m._reset_parameters()
        if not self.two_stage:
            xavier_uniform_(self.reference_points.weight.data, gain=1.0)
            constant_(self.reference_points.bias.data, 0.)
        normal_(self.level_embed)

    def get_proposal_pos_embed(self, proposals):
        num_pos_feats = 128
        temperature = 10000
        scale = 2 * math.pi

        dim_t = torch.arange(num_pos_feats, dtype=torch.float32, device=proposals.device)
        dim_t = temperature ** (2 * (dim_t // 2) / num_pos_feats)
        # N, L, 4
        proposals = proposals.sigmoid() * scale
        # N, L, 4, 128
        pos = proposals[:, :, :, None] / dim_t
        # N, L, 4, 64, 2
        pos = torch.stack((pos[:, :, :, 0::2].sin(), pos[:, :, :, 1::2].cos()), dim=4).flatten(2)
        return pos

    def gen_encoder_output_proposals(self, memory, memory_padding_mask, spatial_shapes):
        N_, S_, C_ = memory.shape
        base_scale = 4.0
        proposals = []
        _cur = 0
        level_ids = []
        for lvl, (H_, W_) in enumerate(spatial_shapes):
            mask_flatten_ = memory_padding_mask[:, _cur:(_cur + H_ * W_)].view(N_, H_, W_, 1)
            valid_H = torch.sum(~mask_flatten_[:, :, 0, 0], 1)
            valid_W = torch.sum(~mask_flatten_[:, 0, :, 0], 1)

            grid_y, grid_x = torch.meshgrid(torch.linspace(0, H_ - 1, H_, dtype=torch.float32, device=memory.device),
                                            torch.linspace(0, W_ - 1, W_, dtype=torch.float32, device=memory.device))
            grid = torch.cat([grid_x.unsqueeze(-1), grid_y.unsqueeze(-1)], -1)

            scale = torch.cat([valid_W.unsqueeze(-1), valid_H.unsqueeze(-1)], 1).view(N_, 1, 1, 2)
            grid = (grid.unsqueeze(0).expand(N_, -1, -1, -1) + 0.5) / scale
            wh = torch.ones_like(grid) * 0.05 * (2.0 ** lvl)
            proposal = torch.cat((grid, wh), -1).view(N_, -1, 4)
            proposals.append(proposal)
            _cur += (H_ * W_)
            level_ids.append(grid.new_ones(H_ * W_, dtype=torch.long) * lvl)
        output_proposals = torch.cat(proposals, 1)
        output_proposals_valid = ((output_proposals > 0.01) & (output_proposals < 0.99)).all(-1, keepdim=True)
        output_proposals = torch.log(output_proposals / (1 - output_proposals))
        output_proposals = output_proposals.masked_fill(memory_padding_mask.unsqueeze(-1), float('inf'))
        output_proposals = output_proposals.masked_fill(~output_proposals_valid, float('inf'))

        output_memory = memory
        output_memory = output_memory.masked_fill(memory_padding_mask.unsqueeze(-1), float(0))
        output_memory = output_memory.masked_fill(~output_proposals_valid, float(0))
        output_memory = self.enc_output_norm(self.enc_output(output_memory))
        level_ids = torch.cat(level_ids)
        return output_memory, output_proposals, level_ids

    def get_valid_ratio(self, mask):
        _, H, W = mask.shape
        valid_H = torch.sum(~mask[:, :, 0], 1)
        valid_W = torch.sum(~mask[:, 0, :], 1)
        valid_ratio_h = valid_H.float() / H
        valid_ratio_w = valid_W.float() / W
        valid_ratio = torch.stack([valid_ratio_w, valid_ratio_h], -1)
        return valid_ratio

    def forward(self, srcs, masks, pos_embeds, query_embed=None):
        assert self.two_stage or query_embed is not None

        # prepare input for encoder
        src_flatten = []
        mask_flatten = []
        lvl_pos_embed_flatten = []
        spatial_shapes = []
        for lvl, (src, mask, pos_embed) in enumerate(zip(srcs, masks, pos_embeds)):
            bs, c, h, w = src.shape
            spatial_shape = (h, w)
            spatial_shapes.append(spatial_shape)
            src = src.flatten(2).transpose(1, 2)
            mask = mask.flatten(1)
            pos_embed = pos_embed.flatten(2).transpose(1, 2)
            lvl_pos_embed = pos_embed + self.level_embed[lvl].view(1, 1, -1)
            lvl_pos_embed_flatten.append(lvl_pos_embed)
            src_flatten.append(src)
            mask_flatten.append(mask)
        src_flatten = torch.cat(src_flatten, 1)
        mask_flatten = torch.cat(mask_flatten, 1)
        lvl_pos_embed_flatten = torch.cat(lvl_pos_embed_flatten, 1)
        spatial_shapes = torch.as_tensor(spatial_shapes, dtype=torch.long, device=src_flatten.device)
        level_start_index = torch.cat((spatial_shapes.new_zeros((1, )), spatial_shapes.prod(1).cumsum(0)[:-1]))
        valid_ratios = torch.stack([self.get_valid_ratio(m) for m in masks], 1)

        # encoder
        memory = self.encoder(src_flatten, spatial_shapes, level_start_index, valid_ratios, lvl_pos_embed_flatten, mask_flatten)

        # prepare input for decoder
        bs, _, c = memory.shape
        if self.two_stage:
            output_memory, output_proposals, level_ids = self.gen_encoder_output_proposals(memory, mask_flatten, spatial_shapes)

            # hack implementation for two-stage Deformable DETR
            enc_outputs_class = self.decoder.class_embed[self.decoder.num_layers](output_memory)
            enc_outputs_coord_unact = self.decoder.bbox_embed[self.decoder.num_layers](output_memory) + output_proposals

            topk = self.two_stage_num_proposals
            proposal_logit = enc_outputs_class[..., 0]

            if self.assign_first_stage:
                proposal_boxes = box_cxcywh_to_xyxy(enc_outputs_coord_unact.sigmoid().float()).clamp(0, 1)
                topk_proposals = []
                for b in range(bs):
                    prop_boxes_b = proposal_boxes[b]
                    prop_logits_b = proposal_logit[b]

                    # pre-nms per-level topk
                    pre_nms_topk = 1000
                    pre_nms_inds = []
                    for lvl in range(len(spatial_shapes)):
                        lvl_mask = level_ids == lvl
                        pre_nms_inds.append(torch.topk(prop_logits_b.sigmoid() * lvl_mask, pre_nms_topk)[1])
                    pre_nms_inds = torch.cat(pre_nms_inds)

                    # nms on topk indices
                    post_nms_inds = batched_nms(prop_boxes_b[pre_nms_inds], prop_logits_b[pre_nms_inds], level_ids[pre_nms_inds], 0.9)
                    keep_inds = pre_nms_inds[post_nms_inds]

                    if len(keep_inds) < self.two_stage_num_proposals:
                        print(f'[WARNING] nms proposals ({len(keep_inds)}) < {self.two_stage_num_proposals}, running naive topk')
                        keep_inds = torch.topk(proposal_logit[b], topk)[1]

                    # keep top Q/L indices for L levels
                    q_per_l = topk // len(spatial_shapes)
                    is_level_ordered = level_ids[keep_inds][None] == torch.arange(len(spatial_shapes), device=level_ids.device)[:,None]  # LS
                    keep_inds_mask = is_level_ordered & (is_level_ordered.cumsum(1) <= q_per_l)  # LS
                    keep_inds_mask = keep_inds_mask.any(0)  # S

                    # pad to Q indices (might let ones filtered from pre-nms sneak by... unlikely because we pick high conf anyways)
                    if keep_inds_mask.sum() < topk:
                        num_to_add = topk - keep_inds_mask.sum()
                        pad_inds = (~keep_inds_mask).nonzero()[:num_to_add]
                        keep_inds_mask[pad_inds] = True

                    # index
                    keep_inds_topk = keep_inds[keep_inds_mask]
                    topk_proposals.append(keep_inds_topk)
                topk_proposals = torch.stack(topk_proposals)
            else:
                topk_proposals = torch.topk(proposal_logit, topk, dim=1)[1]

            topk_coords_unact = torch.gather(enc_outputs_coord_unact, 1, topk_proposals.unsqueeze(-1).repeat(1, 1, 4))
            topk_coords_unact = topk_coords_unact.detach()
            reference_points = topk_coords_unact.sigmoid()
            init_reference_out = reference_points
            pos_trans_out = self.pos_trans_norm(self.pos_trans(self.get_proposal_pos_embed(topk_coords_unact)))
            query_embed, tgt = torch.split(pos_trans_out, c, dim=2)

            topk_feats = torch.stack([output_memory[b][topk_proposals[b]] for b in range(bs)]).detach()
            tgt = tgt + self.pix_trans_norm(self.pix_trans(topk_feats))
        else:
            query_embed, tgt = torch.split(query_embed, c, dim=1)
            query_embed = query_embed.unsqueeze(0).expand(bs, -1, -1)
            tgt = tgt.unsqueeze(0).expand(bs, -1, -1)
            reference_points = self.reference_points(query_embed).sigmoid()
            init_reference_out = reference_points

        # decoder
        hs, inter_references = self.decoder(tgt, reference_points, memory,
                                            spatial_shapes, level_start_index, valid_ratios, query_embed, mask_flatten)

        inter_references_out = inter_references
        if self.two_stage:
            return hs, init_reference_out, inter_references_out, enc_outputs_class, enc_outputs_coord_unact, output_proposals.sigmoid()
        return hs, init_reference_out, inter_references_out, None, None, None


class DeformableTransformerEncoderLayer(nn.Module):
    def __init__(self,
                 d_model=256, d_ffn=1024,
                 dropout=0.1, activation="relu",
                 n_levels=4, n_heads=8, n_points=4):
        super().__init__()

        # self attention
        self.self_attn = MSDeformAttn(d_model, n_levels, n_heads, n_points)
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # ffn
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.activation = _get_activation_fn(activation)
        self.dropout2 = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.dropout3 = nn.Dropout(dropout)
        self.norm2 = nn.LayerNorm(d_model)

    @staticmethod
    def with_pos_embed(tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_ffn(self, src):
        src2 = self.linear2(self.dropout2(self.activation(self.linear1(src))))
        src = src + self.dropout3(src2)
        src = self.norm2(src)
        return src

    def forward(self, src, pos, reference_points, spatial_shapes, level_start_index, padding_mask=None):
        # self attention
        src2 = self.self_attn(self.with_pos_embed(src, pos), reference_points, src, spatial_shapes, level_start_index, padding_mask)
        src = src + self.dropout1(src2)
        src = self.norm1(src)

        # ffn
        src = self.forward_ffn(src)

        return src


class DeformableTransformerEncoder(nn.Module):
    def __init__(self, encoder_layer, num_layers):
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers

    @staticmethod
    def get_reference_points(spatial_shapes, valid_ratios, device):
        reference_points_list = []
        for lvl, (H_, W_) in enumerate(spatial_shapes):

            ref_y, ref_x = torch.meshgrid(torch.linspace(0.5, H_ - 0.5, H_, dtype=torch.float32, device=device),
                                          torch.linspace(0.5, W_ - 0.5, W_, dtype=torch.float32, device=device))
            ref_y = ref_y.reshape(-1)[None] / (valid_ratios[:, None, lvl, 1] * H_)
            ref_x = ref_x.reshape(-1)[None] / (valid_ratios[:, None, lvl, 0] * W_)
            ref = torch.stack((ref_x, ref_y), -1)
            reference_points_list.append(ref)
        reference_points = torch.cat(reference_points_list, 1)
        reference_points = reference_points[:, :, None] * valid_ratios[:, None]
        return reference_points

    def forward(self, src, spatial_shapes, level_start_index, valid_ratios, pos=None, padding_mask=None):
        output = src
        reference_points = self.get_reference_points(spatial_shapes, valid_ratios, device=src.device)
        for _, layer in enumerate(self.layers):
            output = layer(output, pos, reference_points, spatial_shapes, level_start_index, padding_mask)

        return output


class DeformableTransformerDecoderLayer(nn.Module):
    def __init__(self, d_model=256, d_ffn=1024,
                 dropout=0.1, activation="relu",
                 n_levels=4, n_heads=8, n_points=4):
        super().__init__()

        # cross attention
        self.cross_attn = MSDeformAttn(d_model, n_levels, n_heads, n_points)
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # self attention
        self.self_attn = nn.MultiheadAttention(d_model, n_heads, dropout=dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.norm2 = nn.LayerNorm(d_model)

        # ffn
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.activation = _get_activation_fn(activation)
        self.dropout3 = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.dropout4 = nn.Dropout(dropout)
        self.norm3 = nn.LayerNorm(d_model)

    @staticmethod
    def with_pos_embed(tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_ffn(self, tgt):
        tgt2 = self.linear2(self.dropout3(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout4(tgt2)
        tgt = self.norm3(tgt)
        return tgt

    def forward(self, tgt, query_pos, reference_points, src, src_spatial_shapes, level_start_index, src_padding_mask=None):
        # self attention
        q = k = self.with_pos_embed(tgt, query_pos)
        tgt2 = self.self_attn(q.transpose(0, 1), k.transpose(0, 1), tgt.transpose(0, 1))[0].transpose(0, 1)
        tgt = tgt + self.dropout2(tgt2)
        tgt = self.norm2(tgt)

        # cross attention
        tgt2 = self.cross_attn(self.with_pos_embed(tgt, query_pos),
                               reference_points,
                               src, src_spatial_shapes, level_start_index, src_padding_mask)
        tgt = tgt + self.dropout1(tgt2)
        tgt = self.norm1(tgt)

        # ffn
        tgt = self.forward_ffn(tgt)

        return tgt


class DeformableTransformerDecoder(nn.Module):
    def __init__(self, decoder_layer, num_layers, return_intermediate=False):
        super().__init__()
        self.layers = _get_clones(decoder_layer, num_layers)
        self.num_layers = num_layers
        self.return_intermediate = return_intermediate
        # hack implementation for iterative bounding box refinement and two-stage Deformable DETR
        self.bbox_embed = None
        self.class_embed = None

    def forward(self, tgt, reference_points, src, src_spatial_shapes, src_level_start_index, src_valid_ratios,
                query_pos=None, src_padding_mask=None):
        output = tgt

        intermediate = []
        intermediate_reference_points = []
        for lid, layer in enumerate(self.layers):
            if reference_points.shape[-1] == 4:
                reference_points_input = reference_points[:, :, None] \
                                         * torch.cat([src_valid_ratios, src_valid_ratios], -1)[:, None]
            else:
                assert reference_points.shape[-1] == 2
                reference_points_input = reference_points[:, :, None] * src_valid_ratios[:, None]
            output = layer(output, query_pos, reference_points_input, src, src_spatial_shapes, src_level_start_index, src_padding_mask)

            # hack implementation for iterative bounding box refinement
            if self.bbox_embed is not None:
                tmp = self.bbox_embed[lid](output)
                if reference_points.shape[-1] == 4:
                    new_reference_points = tmp + inverse_sigmoid(reference_points)
                    new_reference_points = new_reference_points.sigmoid()
                else:
                    assert reference_points.shape[-1] == 2
                    new_reference_points = tmp
                    new_reference_points[..., :2] = tmp[..., :2] + inverse_sigmoid(reference_points)
                    new_reference_points = new_reference_points.sigmoid()
                reference_points = new_reference_points.detach()

            if self.return_intermediate:
                intermediate.append(output)
                intermediate_reference_points.append(reference_points)

        if self.return_intermediate:
            return torch.stack(intermediate), torch.stack(intermediate_reference_points)

        return output, reference_points


def _get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])


def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    if activation == "glu":
        return F.glu
    raise RuntimeError(F"activation should be relu/gelu, not {activation}.")


def build_deforamble_transformer(args):
    return DeformableTransformer(
        d_model=args.hidden_dim,
        nhead=args.nheads,
        num_encoder_layers=args.enc_layers,
        num_decoder_layers=args.dec_layers,
        dim_feedforward=args.dim_feedforward,
        dropout=args.dropout,
        activation="relu",
        return_intermediate_dec=True,
        num_feature_levels=args.num_feature_levels,
        dec_n_points=args.dec_n_points,
        enc_n_points=args.enc_n_points,
        two_stage=args.two_stage,
        two_stage_num_proposals=args.num_queries,
        assign_first_stage=args.assign_first_stage,
        )





# --------------------------------------------------------
# BEATs: Audio Pre-Training with Acoustic Tokenizers (https://arxiv.org/abs/2212.09058)
# Github source: https://github.com/microsoft/unilm/tree/master/beats
# Copyright (c) 2022 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------


import torch
import torch.nn as nn
from torch.nn import LayerNorm
import torchaudio.compliance.kaldi as ta_kaldi

try:
    from .backbone import (
        TransformerEncoder,
    )
except:
    from backbone import (
        TransformerEncoder,
    )

import logging
from typing import Optional

logger = logging.getLogger(__name__)


class BEATsConfig:
    def __init__(self, cfg=None):
        self.input_patch_size: int = -1  # path size of patch embedding
        self.embed_dim: int = 512  # patch embedding dimension
        self.conv_bias: bool = False  # include bias in conv encoder

        self.encoder_layers: int = 12  # num encoder layers in the transformer
        self.encoder_embed_dim: int = 768  # encoder embedding dimension
        self.encoder_ffn_embed_dim: int = 3072  # encoder embedding dimension for FFN
        self.encoder_attention_heads: int = 12  # num encoder attention heads
        self.activation_fn: str = "gelu"  # activation function to use

        self.layer_wise_gradient_decay_ratio: float = 1.0  # ratio for layer-wise gradient decay
        self.layer_norm_first: bool = False  # apply layernorm first in the transformer
        self.deep_norm: bool = False  # apply deep_norm first in the transformer

        # dropouts
        self.dropout: float = 0.1  # dropout probability for the transformer
        self.attention_dropout: float = 0.1  # dropout probability for attention weights
        self.activation_dropout: float = 0.0  # dropout probability after activation in FFN
        self.encoder_layerdrop: float = 0.0  # probability of dropping a tarnsformer layer
        self.dropout_input: float = 0.0  # dropout to apply to the input (after feat extr)

        # positional embeddings
        self.conv_pos: int = 128  # number of filters for convolutional positional embeddings
        self.conv_pos_groups: int = 16  # number of groups for convolutional positional embedding

        # relative position embedding
        self.relative_position_embedding: bool = False  # apply relative position embedding
        self.num_buckets: int = 320  # number of buckets for relative position embedding
        self.max_distance: int = 1280  # maximum distance for relative position embedding
        self.gru_rel_pos: bool = False  # apply gated relative position embedding

        # label predictor
        self.finetuned_model: bool = False  # whether the model is a fine-tuned model.
        self.predictor_dropout: float = 0.1  # dropout probability for the predictor
        self.predictor_class: int = 527  # target class number for the predictor

        if cfg is not None:
            self.update(cfg)

    def update(self, cfg: dict):
        self.__dict__.update(cfg)


class BEATs(nn.Module):
    def __init__(
            self,
            cfg: BEATsConfig,
    ) -> None:
        super().__init__()
        logger.info(f"BEATs Config: {cfg.__dict__}")

        self.cfg = cfg

        self.embed = cfg.embed_dim
        self.post_extract_proj = (
            nn.Linear(self.embed, cfg.encoder_embed_dim)
            if self.embed != cfg.encoder_embed_dim
            else None
        )

        self.input_patch_size = cfg.input_patch_size
        self.patch_embedding = nn.Conv2d(1, self.embed, kernel_size=self.input_patch_size, stride=self.input_patch_size,
                                         bias=cfg.conv_bias)

        self.dropout_input = nn.Dropout(cfg.dropout_input)

        assert not cfg.deep_norm or not cfg.layer_norm_first
        self.encoder = TransformerEncoder(cfg)
        self.layer_norm = LayerNorm(self.embed)

        if cfg.finetuned_model:
            self.predictor_dropout = nn.Dropout(cfg.predictor_dropout)
            self.predictor = nn.Linear(cfg.encoder_embed_dim, cfg.predictor_class)
        else:
            self.predictor = None

    def forward_padding_mask(
            self,
            features: torch.Tensor,
            padding_mask: torch.Tensor,
    ) -> torch.Tensor:
        extra = padding_mask.size(1) % features.size(1)
        if extra > 0:
            padding_mask = padding_mask[:, :-extra]
        padding_mask = padding_mask.view(
            padding_mask.size(0), features.size(1), -1
        )
        padding_mask = padding_mask.all(-1)
        return padding_mask

    def preprocess(
            self,
            source: torch.Tensor,
            fbank_mean: float = 15.41663,
            fbank_std: float = 6.55582,
    ) -> torch.Tensor:
        fbanks = []
        for waveform in source:
            waveform = waveform.unsqueeze(0) * 2 ** 15
            # print(waveform.max(), waveform.min(), waveform.shape, waveform.dtype)
            # waveform = waveform.unsqueeze(0)
            fbank = ta_kaldi.fbank(waveform, num_mel_bins=128, sample_frequency=16000, frame_length=25, frame_shift=10)
            # print(fbank.max(), fbank.min(), fbank.shape, fbank.dtype)
            fbanks.append(fbank)
        fbank = torch.stack(fbanks, dim=0)
        fbank = (fbank - fbank_mean) / (2 * fbank_std)
        # print(fbank.max(), fbank.min(), fbank.shape, fbank.dtype)
        return fbank

    def forward(self, fbank):
        fbank = fbank.unsqueeze(1)
        ### fbank b,1,128,1024
        features = self.patch_embedding(fbank)
        ### b,512,8,64
        features = features.reshape(features.shape[0], features.shape[1], -1).contiguous()
        ### b, 512 , 512
        features = features.transpose(1, 2).contiguous() ##b,512,512
        features = self.layer_norm(features)
        x = self.dropout_input(features)

        if self.post_extract_proj is not None:
            features = self.post_extract_proj(features)

        x = self.dropout_input(features)
        x, layer_results = self.encoder(
            x,
            padding_mask=None,
        )

        return x 

    def extract_features(
            self,
            source: torch.Tensor,
            padding_mask: Optional[torch.Tensor] = None,
            fbank_mean: float = 15.41663,
            fbank_std: float = 6.55582,
    ):
        fbank = self.preprocess(source, fbank_mean=fbank_mean, fbank_std=fbank_std)

        if padding_mask is not None:
            padding_mask = self.forward_padding_mask(fbank, padding_mask)

        fbank = fbank.unsqueeze(1)
        features = self.patch_embedding(fbank)
        features = features.reshape(features.shape[0], features.shape[1], -1)
        features = features.transpose(1, 2).contiguous()
        features = self.layer_norm(features)

        if padding_mask is not None:
            padding_mask = self.forward_padding_mask(features, padding_mask)

        if self.post_extract_proj is not None:
            features = self.post_extract_proj(features)

        x = self.dropout_input(features)

        x, layer_results = self.encoder(
            x,
            padding_mask=padding_mask,
        )

        if self.predictor is not None:
            x = self.predictor_dropout(x)
            logits = self.predictor(x)

            if padding_mask is not None and padding_mask.any():
                logits[padding_mask] = 0
                logits = logits.sum(dim=1)
                logits = logits / (~padding_mask).sum(dim=1).unsqueeze(-1).expand_as(logits)
            else:
                logits = logits.mean(dim=1)

            lprobs = torch.sigmoid(logits)

            return lprobs, padding_mask
        else:
            return x, padding_mask



"""Video type classifier module.

From NVIDIA Curator's video type classification concept.
Classifies content as real-world, animated, game, abstract, etc.
Uses CLIP zero-shot classification.
"""

import logging
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

VIDEO_TYPE_LABELS = [
    "real-world photography",
    "animated cartoon or anime",
    "video game footage",
    "abstract visual pattern",
    "screen recording or presentation",
    "text-heavy or document",
]

VIDEO_TYPE_SHORT = [
    "real",
    "animated",
    "game",
    "abstract",
    "screen_recording",
    "text_heavy",
]


class VideoTypeClassifierModule(PipelineModule):
    name = "video_type_classifier"
    description = "CLIP zero-shot video content type classification"
    default_config = {"subsample": 4}

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._ml_available = False
        self._model = None
        self._processor = None
        self._device = None

    def setup(self) -> None:
        try:
            import torch
            from transformers import CLIPModel, CLIPProcessor

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            model_name = "openai/clip-vit-base-patch32"
            self._model = CLIPModel.from_pretrained(model_name).to(device)
            self._processor = CLIPProcessor.from_pretrained(model_name)
            self._model.eval()
            self._device = device
            self._ml_available = True
            logger.info("CLIP model loaded for video type classification on %s", device)
        except (ImportError, Exception) as e:
            logger.warning("Video type classifier unavailable: %s", e)

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        if not self._ml_available:
            return sample

        try:
            import cv2
            import torch
            from PIL import Image

            subsample = self.config.get("subsample", 4)
            frames = []

            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                indices = list(range(0, total, max(1, total // subsample)))[:subsample]
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                        frames.append(Image.fromarray(rgb))
                cap.release()
            else:
                img = Image.open(str(sample.path))
                try:
                    frames.append(img.convert("RGB"))
                except Exception:
                    img.close()
                    raise

            if not frames:
                return sample

            # Classify each frame and aggregate
            votes = np.zeros(len(VIDEO_TYPE_LABELS))
            for img in frames:
                inputs = self._processor(
                    text=VIDEO_TYPE_LABELS,
                    images=img,
                    return_tensors="pt",
                    padding=True,
                ).to(self._device)

                with torch.no_grad():
                    outputs = self._model(**inputs)
                    logits = outputs.logits_per_image.squeeze()
                    probs = torch.softmax(logits, dim=-1).cpu().numpy()
                    votes += probs

            votes /= len(frames)
            best_idx = int(np.argmax(votes))
            sample.quality_metrics.video_type = VIDEO_TYPE_SHORT[best_idx]
            sample.quality_metrics.video_type_confidence = float(votes[best_idx])
        except Exception as e:
            logger.warning("Video type classification failed: %s", e)
        return sample



# Copyright (c) Meta Platforms, Inc. and affiliates.

import re
from dataclasses import dataclass
from logging import getLogger
from typing import Any, Dict, List, Optional

import torch

from ...core.distributed import get_is_master

logger = getLogger()


@dataclass
class MLLMBatch:
    x: torch.LongTensor
    y: torch.LongTensor
    mask: Optional[torch.BoolTensor] = None
    image_pos_index: Optional[torch.LongTensor] = None
    images: Optional[torch.Tensor] = None
    media_type: Optional[List[str]] = (["text"],)
    num_image_chunks: Optional[List[int]] = None

    def __post_init__(self):
        assert self.x.dim() == 2, "{} != 2".format(self.x.dim())
        assert self.x.shape == self.y.shape
        assert self.x.dtype == torch.int64
        assert self.y.dtype == torch.int64
        assert self.mask is None or self.mask.shape == self.x.shape


class BaseCollator:
    def __init__(
        self,
        tokenizer,
        show_first_batch: bool = False,
    ) -> None:
        self.tokenizer = tokenizer
        self.first_batch = show_first_batch

    def __call__(self, features: List[Dict[str, Any]]):
        raise NotImplementedError


class MllmPaddingCollator(BaseCollator):

    def prettify_decoded_text(self, texts: List[str]) -> List[str]:
        """
        Prettify the decoded text by replacing consecutive <|image|> tokens with a shortened form using regex.
        """
        prettified = []
        special_tokens = ["<|end_of_text|>", "<|image|>"]
        for text in texts:
            for token in special_tokens:
                # Regex to find consecutive occurrences of the token
                pattern = f"({re.escape(token)})\\1+"  # Captures repeating groups of the token

                def replace_consecutive(match):
                    count = len(match.group(0)) // len(token)
                    return f"{token}..x{count}"

                text = re.sub(pattern, replace_consecutive, text)
            prettified.append(text)
        return prettified

    def __call__(self, features: List[Dict[str, Any]]) -> MLLMBatch:
        text = []
        images = []
        media_type = []
        response_pos = []
        image_pos = []
        num_image_chunks = []
        for b in features:
            text.append(b["text_ids"])
            images.append(b["media"])
            response_pos.append(b["response_pos"])
            image_pos.append(b["image_pos"])
            num_image_chunks.append(b["num_image_chunks"])
            media_type.append(b["media_type"])

        images = [img for img in images if img is not None]
        images = torch.cat(images) if images else None

        # max_text_len = max([len(x) for x in text]) - 1
        bsz = len(text)
        input_ids = torch.full(
            (bsz, self.tokenizer.seq_len), self.tokenizer.pad_token_id
        )
        label_ids = torch.full(
            (bsz, self.tokenizer.seq_len), self.tokenizer.pad_token_id
        )
        image_pos_index = torch.full((bsz, self.tokenizer.seq_len), -1)

        for i in range(bsz):
            # Shift labels (list of lists) to train next token prediction
            for j in response_pos[i]:
                label_ids[i][j - 1] = text[i][j]
            # Remove last token for input
            text_len = len(text[i]) - 1
            input_ids[i][:text_len] = torch.tensor(text[i][:-1])
            # Fill image_pos_index
            if image_pos[i]:
                image_indices = torch.arange(len(image_pos[i]))
                image_pos_index[i, image_pos[i]] = image_indices

        mask = label_ids.ne(self.tokenizer.pad_token_id)

        # Replace all pad tokens with eos tokens
        input_ids[input_ids == self.tokenizer.pad_token_id] = (
            self.tokenizer.eos_token_id
        )
        label_ids[label_ids == self.tokenizer.pad_token_id] = (
            self.tokenizer.eos_token_id
        )

        if self.first_batch and get_is_master():
            input_decoded = self.tokenizer.decode_batch(input_ids)
            label_decoded = self.tokenizer.decode_batch(label_ids)
            logger.info(f"Input text: \n{self.prettify_decoded_text(input_decoded)}")
            logger.info(f"Label text: \n{self.prettify_decoded_text(label_decoded)}")
            self.first_batch = False

        return MLLMBatch(
            x=input_ids,
            y=label_ids,
            mask=mask,
            image_pos_index=image_pos_index,
            images=images,
            media_type=media_type,
            num_image_chunks=num_image_chunks,
        )



import torch
import torch.nn as nn
import re


class IdentityMap(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, *args, **kwargs):
        return x

    @property
    def config(self):
        return {"mm_projector_type": 'identity'}


class SimpleResBlock(nn.Module):
    def __init__(self, channels):
        super().__init__()
        self.pre_norm = nn.LayerNorm(channels)

        self.proj = nn.Sequential(
            nn.Linear(channels, channels),
            nn.GELU(),
            nn.Linear(channels, channels)
        )
    def forward(self, x):
        x = self.pre_norm(x)
        return x + self.proj(x)


def build_vision_projector(config, delay_load=False, **kwargs):
    projector_type = getattr(config, 'mm_projector_type', 'linear')

    if projector_type == 'linear':
        return nn.Linear(config.mm_hidden_size, config.hidden_size)

    mlp_gelu_match = re.match(r'^mlp(\d+)x_gelu$', projector_type)
    if mlp_gelu_match:
        mlp_depth = int(mlp_gelu_match.group(1))
        modules = [nn.Linear(config.mm_hidden_size, config.hidden_size)]
        for _ in range(1, mlp_depth):
            modules.append(nn.GELU())
            modules.append(nn.Linear(config.hidden_size, config.hidden_size))
        return nn.Sequential(*modules)

    if projector_type == 'identity':
        return IdentityMap()

    raise ValueError(f'Unknown projector type: {projector_type}')



import os
from .clip_encoder import CLIPVisionTower


def build_vision_tower(vision_tower_cfg, **kwargs):
    vision_tower = getattr(vision_tower_cfg, 'mm_vision_tower', getattr(vision_tower_cfg, 'vision_tower', None))
    is_absolute_path_exists = os.path.exists(vision_tower)
    if is_absolute_path_exists or vision_tower.startswith("openai") or vision_tower.startswith("laion") or "ShareGPT4V" in vision_tower:
        return CLIPVisionTower(vision_tower, args=vision_tower_cfg, **kwargs)

    raise ValueError(f'Unknown vision tower: {vision_tower}')


import datetime
import logging
import time
from os.path import join

import pandas as pd
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import wandb

from dataset.serialize import local_broadcast_process_authkey
from dataset import MetaLoader, create_dataset, create_loader, create_sampler, create_stateful_sampler
from models import *
from tasks.retrieval_utils import evaluation_wrapper
from tasks.shared_utils import get_media_types, setup_model
from utils.basic_utils import (MetricLogger, SmoothedValue,
                               remove_files_if_exist, setup_seed)
from utils.config_utils import setup_main
from utils.distributed import get_rank, get_world_size, is_main_process
from utils.logger import log_dict_to_wandb, setup_wandb
try:
    from petrel_client.client import Client
except:
    Client = None
import io
import os
import shutil

logger = logging.getLogger(__name__)

ceph_ckpt_bucket = "shdd:s3://avp_ckpt"


def train(
    model,
    train_loaders,
    optimizer,
    tokenizer,
    epoch,
    global_step,
    device,
    scheduler,
    scaler,
    config,
    skip_num=0
):

    try:
        ceph_ckpt_path = f"{ceph_ckpt_bucket}/{config.output_dir.split('/')[-3]}/{config.output_dir.split('/')[-2]}/{config.output_dir.split('/')[-1]}"
        client_ckpt = Client(conf_path='~/petreloss.conf')
    except Exception as e:
        print(e)
        logger.info("Ceph is not working!!!")


    if config.use_half_precision: 
        if config.get('use_bf16', False):
            cast_dtype = torch.bfloat16
        else:
            cast_dtype = torch.float16
    else:
        cast_dtype = None

    model.train()

    metric_logger = MetricLogger(delimiter="  ")
    metric_logger.add_meter("lr", SmoothedValue(window=100, fmt="{value:.6f}"))
    metric_logger.add_meter("temperature", SmoothedValue(window=100, fmt="{value:.4f}"))
    loss_names = ["loss_" + k for k, v in config.criterion.loss_weight.items() if v != 0]
    
    if config.get("use_raw_text", False): # for cosa
        loss_names = loss_names + ["c_loss_" + k for k, v in config.criterion.loss_weight.items() if v != 0]
    uta_all = config.criterion.get('uta_all', False)

    media_types = get_media_types(train_loaders)

    for name in loss_names:
        for m in media_types:
            metric_logger.add_meter(
                f"{m}-{name}", SmoothedValue(window=100, fmt="{value:.4f}")
            )

    header = f"Train Epoch: [{epoch}]"
    log_freq = config.log_freq

    if config.distributed:
        for d in train_loaders:
            d.sampler.set_epoch(epoch)

    if config.get('use_iter_train', False):
        train_loader = MetaLoader_rs2(name2loader=dict(list(zip(media_types, train_loaders))), skip_num=skip_num)
    else:
        train_loader = MetaLoader(name2loader=dict(list(zip(media_types, train_loaders))))

    model_without_ddp = model.module if config.distributed else model
    iterator = metric_logger.log_every(train_loader, log_freq, header)
    
    begin_step = global_step % len(train_loader)
    logger.info(f"Epoch={epoch}, begin_step={begin_step} save_ckpt_iter={config.get('save_ckpt_iter', None)}")

    for local_step, (media_type, (media, text, idx)) in enumerate(iterator):
        if local_step < begin_step: 
            logger.warn(f"Jump local_step: {local_step} (begin_step={begin_step})!!!")
            metric_logger.update(lr=optimizer.param_groups[0]["lr"])
            continue

        if config.get("save_ckpt_iter", None) is not None: # and not is_iter_resume:
            if local_step != 0 and local_step % config.get("save_ckpt_iter") == 0:
                if hasattr(config, "deepspeed") and config.deepspeed.enable: 
                    tag = f"ckpt_e{epoch:02d}_local{local_step}_global{global_step}"
                    client_state = {'epoch': epoch, 'global_step': global_step, 'local_step': local_step}
                    model.save_checkpoint(config.output_dir, tag=tag, save_latest=False, client_state=client_state)
                    logger.info(f"save ckpt file to local ({config.output_dir}/{tag})!!!")
                elif is_main_process():
                    state_dict = model_without_ddp.state_dict()
                    for k in config.get("no_save_params_prefix", []):
                        kk = [x for x in state_dict.keys() if x.startswith(k)]
                        logger.info(f"Not saving {len(kk)} params with prefix {k}")
                        for kkk in kk:
                            state_dict.pop(kkk)

                    save_obj = {
                        "model": state_dict,
                        "optimizer": optimizer.state_dict(),
                        "scheduler": scheduler.state_dict(),
                        "scaler": scaler.state_dict(),
                        "config": config,
                        "epoch": epoch,
                        "local_step": local_step,
                        "global_step": global_step,
                    }
                    try:
                        with io.BytesIO() as buffer:
                            torch.save(save_obj, buffer)
                            client_ckpt.put(f"{ceph_ckpt_path}/ckpt_{epoch:02d}_local{local_step}_global{global_step}.pth", buffer.getvalue())
                            logger.info(f"Save to ceph ({ceph_ckpt_path}/ckpt_{epoch:02d}_local{local_step}_global{global_step}.pth)!!!")
                    except Exception as e:
                        print(e)
                        torch.save(save_obj, join(config.output_dir, f"ckpt_{epoch:02d}_local{local_step}_global{global_step}.pth"))
                        logger.warn(f"Ceph is not working, save to local ({join(config.output_dir, f'ckpt_{epoch:02d}_local{local_step}_global{global_step}.pth')})!!!")

        if media_type == 'audio_video':
            if type(media[0]) is list:
                assert len(media[0]) == 2
                audio = [media[0][0].to(device, dtype=cast_dtype, non_blocking=True), media[0][1].to(device, non_blocking=True)]
            else:
                audio = media[0].to(device, dtype=cast_dtype, non_blocking=True)
            video = media[1].to(device, dtype=cast_dtype, non_blocking=True)
            media = [audio, video]
        else:
            media = media.to(device, dtype=cast_dtype, non_blocking=True)
        idx = idx.to(device, non_blocking=True)
        if config.get("use_raw_text", False) or config.get("use_cosa", False):
            max_length = config.inputs.max_txt_l[media_type]
        else:
            if type(text) is dict:
                text_input = {}
                for k in text.keys():
                    text_input[k] = tokenizer(
                    text[k],
                    padding="max_length",
                    truncation=True,
                    max_length=config.inputs.max_txt_l[media_type],
                    return_tensors="pt",
                ).to(
                    device) # change from "longest" to "max_length"
            else:
                text_input = tokenizer(
                    text,
                    padding="max_length",
                    truncation=True,
                    max_length=config.inputs.max_txt_l[media_type],
                    return_tensors="pt",
                ).to(
                    device) # change from "longest" to "max_length"
            

        if hasattr(config, "deepspeed") and config.deepspeed.enable:
            loss_dict = model(media, text_input, idx=idx, media_type=media_type)
            loss = sum(loss_dict.values())

            model.backward(loss)
            model.step()

        else: # NOTE We shouldn't use scaler if we only involve bf16, check this!
            with torch.cuda.amp.autocast(enabled=config.use_half_precision, dtype=cast_dtype):
                loss_dict = model(media, text_input, idx=idx, media_type=media_type)
                loss = sum(loss_dict.values())

            if not config.use_half_precision or config.get('use_bf16', False):
                optimizer.zero_grad()
                loss.backward()
                if config.optimizer.max_grad_norm > 0:
                    torch.nn.utils.clip_grad_norm_(model.parameters(), config.optimizer.max_grad_norm)
                    
                optimizer.step()
                scheduler.step()
            else:
                optimizer.zero_grad()
                scaler.scale(loss).backward()
                if config.optimizer.max_grad_norm > 0:
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(model.parameters(), config.optimizer.max_grad_norm)
                scaler.step(optimizer)
                scaler.update()
                scheduler.step()

        # logging
        for name in loss_names:
            if name in loss_dict.keys():
                value = loss_dict[name]
                value = value if isinstance(value, float) else value.item()
                metric_logger.update(**{f"{media_type}-{name}": value})
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])
        metric_logger.update(temperature=model_without_ddp.temp.item())

        if is_main_process() and config.wandb.enable and global_step % log_freq == 0:
            try:
                logs = metric_logger.get_global_avg_dict()
                log_dict_to_wandb(logs, step=global_step, prefix="train/")
            except Exception as e:
                logger.warn("Wandb is not working!!!")
                print(e)
            
        global_step += 1

        if config.debug and global_step % 20 == 0:
            logger.info("debug mode, break training loop")
            break

        if config.debug and global_step % (2 * log_freq + 3) == 0:
            logger.info("debug mode, break training loop")
            break

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    logger.info(f"Averaged stats: {metric_logger.global_avg()}")
    return global_step


def setup_dataloaders(config, mode="pt"):
    # train datasets, create a list of data loaders
    logger.info(f"Creating dataset for {mode} use_iter_train={config.get('use_iter_train', False)}")
    train_datasets = create_dataset(f"{mode}_train", config)
    media_types = get_media_types(train_datasets)

    if config.get('use_iter_train', False):
        if config.distributed:
            batch_size = [config.inputs.batch_size[k] for k in media_types] # batch_size for each GPU
            samplers = create_stateful_sampler(train_datasets, batch_size)
        else:
            raise NotImplementedError
    else:
        if config.distributed:
            num_tasks = get_world_size()
            global_rank = get_rank()
            samplers = create_sampler(
                train_datasets, [True] * len(media_types), num_tasks, global_rank
            )
        else:
            samplers = [None] * len(media_types)

    train_loaders = create_loader(
        train_datasets,
        samplers,
        batch_size=[config.inputs.batch_size[k] for k in media_types],
        num_workers=[config.num_workers] * len(media_types),
        is_trains=[True] * len(media_types),
        collate_fns=[None] * len(media_types),
    )  # [0]

    # test datasets, a mapping from dataset name to data loader
    test_datasets, test_dataset_names = create_dataset(f"{mode}_eval", config)
    test_loaders = create_loader(
        test_datasets,
        [None] * len(test_datasets),
        batch_size=[config.inputs.batch_size_test[d.media_type] for d in test_datasets],
        num_workers=[config.num_workers] * len(test_datasets),
        is_trains=[False] * len(test_datasets),
        collate_fns=[None] * len(test_datasets),
    )
    test_name2loaders = {k: v for k, v in zip(test_dataset_names, test_loaders)}
    return train_loaders, test_name2loaders, media_types


def main(config):
    if config.get('use_flash_sdp', False):
        torch.backends.cuda.enable_flash_sdp(enabled=True)
    elif config.get('use_mem_efficient_sdp', False):
        torch.backends.cuda.enable_mem_efficient_sdp(enabled=True)

    try:
        ceph_ckpt_path = f"{ceph_ckpt_bucket}/{config.output_dir.split('/')[-3]}/{config.output_dir.split('/')[-2]}/{config.output_dir.split('/')[-1]}"
        client_ckpt = Client(conf_path='~/petreloss.conf')
    except Exception as e:
        print(e)
        logger.info("Ceph is not working!!!")
        
    if is_main_process() and config.wandb.enable:
        try:
            run = setup_wandb(config)
            logger.info("Wandb is working!!!")
        except Exception as e:
            logger.warn("Wandb is not working!!!")
            print(e)

    is_pretrain = config.mode == "pt"

    logger.info(f"train_file: {config.train_file}")

    setup_seed(config.seed + get_rank())
    device = torch.device(config.device)

    train_loaders, test_name2loaders, train_media_types = setup_dataloaders(
        config, mode=config.mode
    )
    num_steps_per_epoch = sum(len(d) for d in train_loaders)
    config.scheduler.num_training_steps = num_steps_per_epoch * config.scheduler.epochs
    config.scheduler.num_warmup_steps = num_steps_per_epoch * config.scheduler.warmup_epochs
    # set cudnn.benchmark=True only when input size is fixed
    # https://discuss.pytorch.org/t/what-does-torch-backends-cudnn-benchmark-do/5936/3
    cudnn.benchmark = len(train_media_types) == 1

    print(f"\033[31m CURRENT NODE NAME: {os.environ['SLURMD_NODENAME']} dataloader is OK {datetime.datetime.now().strftime('%Y-%m-%d-%H_%M_%S')}!!! \033[0m")

    find_unused_parameters = config.model.get('find_unused_parameters', False)
    logger.info(f"find_unused_parameters={find_unused_parameters}")

    model_cls = eval(config.model.get('model_cls'))
    (
        model,
        model_without_ddp,
        optimizer,
        scheduler,
        scaler,
        tokenizer,
        start_epoch,
        global_step,
    ) = setup_model(
        config,
        model_cls=model_cls,
        add_decoder=False,
        pretrain=is_pretrain,
        find_unused_parameters=find_unused_parameters,
    )

    if is_main_process() and config.wandb.enable:
        try:
            wandb.watch(model)
        except Exception as e:
            logger.warn("Wandb is not working!!!")
            print(e)

    best = 0
    best_epoch = 0
    if type(config.best_key) is str:
        best_key = [config.best_key, "t2v_r1"]
    elif type(config.best_key) is list and len(config.best_key) == 2:
        best_key = config.best_key
    else:
        raise NotImplementedError(config.best_key)

    best_ckpt_id = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    logger.info(f"Start training, start_epoch={start_epoch}")
    start_time = time.time()
    start_step = start_epoch * num_steps_per_epoch
    for epoch in range(start_epoch, config.scheduler.epochs):
        if not config.evaluate:
            global_step = train(
                model,
                train_loaders,
                optimizer,
                tokenizer,
                epoch,
                global_step,
                device,
                scheduler,
                scaler,
                config,
                skip_num = global_step - start_step
            )

            if hasattr(config, "deepspeed") and config.deepspeed.enable:
                if config.get("save_latest", False):
                    tag = "ckpt_latest"
                else:
                    tag = f"ckpt_{epoch:02d}"

                client_state = {'epoch': epoch, 'global_step': global_step}
                model.save_checkpoint(config.output_dir, tag=tag, save_latest=False, client_state=client_state)

                logger.info(f"save ckpt file to local ({config.output_dir}/{tag})!!!") 
                if is_main_process() and config.get("delete_ds_optim_states", False):
                    if config.get("save_latest", False):
                        if epoch == (config.scheduler.epochs - 1): # last epoch
                            last_tag = "ckpt_latest"
                            last_ckpt_path = f"{config.output_dir}/{last_tag}"
                            if os.path.exists(last_ckpt_path):
                                logger.info(f"remove optim states in ({config.output_dir}/{last_tag})!!!")
                                for file in os.listdir(last_ckpt_path):
                                    if file.endswith('optim_states.pt'):
                                        os.remove(os.path.join(last_ckpt_path, file))
                    else:
                        last_tag = f"ckpt_{epoch-1:02d}"
                        last_ckpt_path = f"{config.output_dir}/{last_tag}"
                        if os.path.exists(last_ckpt_path):
                            logger.info(f"remove optim states in ({config.output_dir}/{last_tag})!!!")
                            for file in os.listdir(last_ckpt_path):
                                if file.endswith('optim_states.pt'):
                                    os.remove(os.path.join(last_ckpt_path, file))
                        
                        if epoch == (config.scheduler.epochs - 1): # last epoch
                            last_tag = f"ckpt_{epoch:02d}"
                            last_ckpt_path = f"{config.output_dir}/{last_tag}"
                            if os.path.exists(last_ckpt_path):
                                logger.info(f"remove optim states in ({config.output_dir}/{last_tag})!!!")
                                for file in os.listdir(last_ckpt_path):
                                    if file.endswith('optim_states.pt'):
                                        os.remove(os.path.join(last_ckpt_path, file))

            if is_main_process():
                if not (hasattr(config, "deepspeed") and config.deepspeed.enable):
                    state_dict = model_without_ddp.state_dict()
                    for k in config.get("no_save_params_prefix", []):
                        kk = [x for x in state_dict.keys() if x.startswith(k)]
                        logger.info(f"Not saving {len(kk)} params with prefix {k}")
                        for kkk in kk:
                            state_dict.pop(kkk)

                    save_obj = {
                        "model": state_dict,
                        "optimizer": optimizer.state_dict(),
                        "scheduler": scheduler.state_dict(),
                        "scaler": scaler.state_dict(),
                        "config": config,
                        "epoch": epoch,
                        "global_step": global_step,
                    }
                    try:
                        with io.BytesIO() as buffer:
                            torch.save(save_obj, buffer)
                            if config.get("save_latest", False):
                                client_ckpt.put(f"{ceph_ckpt_path}/ckpt_latest.pth", buffer.getvalue())
                                logger.info(f"Save to ceph ({ceph_ckpt_path}/ckpt_latest.pth)!!!")
                            else:
                                client_ckpt.put(f"{ceph_ckpt_path}/ckpt_{epoch:02d}.pth", buffer.getvalue())
                                logger.info(f"Save to ceph ({ceph_ckpt_path}/ckpt_{epoch:02d}.pth)!!!")
                    except Exception as e:
                        print(e)
                        if config.get("save_latest", False):
                            torch.save(save_obj, join(config.output_dir, "ckpt_latest.pth"))
                            logger.warn(f"Ceph is not working, save to local ({join(config.output_dir, 'ckpt_latest.pth')})!!!")
                        else:
                            torch.save(save_obj, join(config.output_dir, f"ckpt_{epoch:02d}.pth"))
                            logger.warn(f"Ceph is not working, save to local ({join(config.output_dir, f'ckpt_{epoch:02d}.pth')})!!!")



        if config.get("jump_evaluate", False) and not config.evaluate:
            logger.warn(f"Jump the evaluation'))!!!")
        else:
            try:
                eval_res = {}
                for test_name, test_loader in test_name2loaders.items():
                    if test_name not in config.test_types:
                        logger.info(
                            f"Skip eval {test_name} split. All test_types {config.test_types}"
                        )
                        continue
                    res = evaluation_wrapper(
                        model_without_ddp, test_loader, tokenizer, device, config, prefix=test_name
                    )
                    eval_res.update(res)

                if is_main_process():
                    # log to wandb
                    if config.wandb.enable:
                        try:
                            for p, v in eval_res.items():
                                log_dict_to_wandb(v, step=global_step, prefix=p)
                        except Exception as e:
                            logger.warn("Wandb is not working!!!")
                            print(e)

                    try:
                        cur_recall = eval_res[best_key[0]][best_key[1]]
                    except Exception as e:
                        logger.warn(e)
                        print(e)
                        # print(eval_res)
                        cur_recall = best - 1

                    eval_res = pd.DataFrame(eval_res)
                    logger.info(f"Epoch {epoch}")
                    logger.info(f"\n{eval_res.transpose().to_string(max_cols=30)}")

                    eval_res.to_json(join(config.output_dir, "eval_res_latest.json"))

                    state_dict = model_without_ddp.state_dict()

                    for k in config.get("no_save_params_prefix", []):
                        kk = [x for x in state_dict.keys() if x.startswith(k)]
                        logger.info(f"Not saving {len(kk)} params with prefix {k}")
                        for kkk in kk:
                            state_dict.pop(kkk)

                    if not config.evaluate and cur_recall > best:
                        if not (hasattr(config, "deepspeed") and config.deepspeed.enable):
                            try:
                                with io.BytesIO() as buffer:
                                    torch.save(save_obj, buffer)
                                    client_ckpt.put(f"{ceph_ckpt_path}/ckpt_best_{best_ckpt_id}.pth", buffer.getvalue())
                                    logger.info(f"Save to ceph ({f'{ceph_ckpt_path}/ckpt_best_{best_ckpt_id}.pth'})!!!")
                            except Exception as e:
                                print(e)
                                torch.save(save_obj, join(config.output_dir, f"ckpt_best_{best_ckpt_id}.pth"))
                                logger.warn(f"Ceph is not working, save to local ({join(config.output_dir, f'ckpt_best_{best_ckpt_id}.pth')})!!!")
                        else:
                            
                            now_ckpt_path = f"{config.output_dir}/{tag}/mp_rank_00_model_states.pt"
                            best_ckpt_path = f"{config.output_dir}/best_mp_rank_00_model_states.pt"

                            if os.path.exists(now_ckpt_path):
                                shutil.copy(now_ckpt_path, best_ckpt_path)
                                logger.info(f"Copy {now_ckpt_path} to {best_ckpt_path}!!!")
                            else:
                                logger.warn(f"Can't find {now_ckpt_path}, there's some wrong!!!")

                        eval_file = "eval_res_best.json"
                        eval_res.to_json(join(config.output_dir, eval_file))
                        best = cur_recall
                        best_epoch = epoch
            except Exception as e:
                logger.warn("Something wrong when eval or save!!!")
                print(e)
                if config.evaluate:
                    raise e


        if config.evaluate:
            break
        
        start_step = global_step

        dist.barrier()

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    logger.info(f"Training time {total_time_str}")
    logger.info(f"best epoch {best_epoch} [best_key {best_key}]")
    logger.info(f"Checkpoints and Logs saved at {config.output_dir}")

    if is_main_process() and config.wandb.enable:
        try:
            run.finish()
        except Exception as e:
            logger.warn("Wandb is not working!!!")
            print(e)


if __name__ == "__main__":
    print(f"\033[31m NODE LIST: {os.environ['SLURM_NODELIST']} \033[0m")
    logger.info(f"NODE LIST: {os.environ['SLURM_NODELIST']}")
    cfg = setup_main()
    local_broadcast_process_authkey()
    main(cfg)



"""Tests for decoder_stress module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_decoder_stress_basics():
    from ayase.modules.decoder_stress import DecoderStressModule
    _test_module_basics(DecoderStressModule, "decoder_stress")

def test_decoder_stress_video(video_sample):
    from ayase.modules.decoder_stress import DecoderStressModule
    video_sample.quality_metrics = QualityMetrics()
    m = DecoderStressModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

__author__ = "aagrawal"



from typing import List

from .score import Score

from .constants import HF_CACHE_DIR

from .models.clipscore_models import list_all_clipscore_models, get_clipscore_model

class CLIPScore(Score):
    def prepare_scoremodel(self,
                           model='openai:ViT-L/14',
                           device='cuda',
                           cache_dir=HF_CACHE_DIR,
                           **kwargs):

        return get_clipscore_model(
            model,
            device=device,
            cache_dir=cache_dir,
            **kwargs
        )
            
    def list_all_models(self) -> List[str]:
        return list_all_clipscore_models()


"""Tests for generative_distribution module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_generative_distribution_basics():
    from ayase.modules.generative_distribution_metrics import GenerativeDistributionModule
    _test_module_basics(GenerativeDistributionModule, "generative_distribution")

def test_generative_distribution_extract(video_sample):
    from ayase.modules.generative_distribution_metrics import GenerativeDistributionModule
    m = GenerativeDistributionModule()
    feat = m.extract_features(video_sample)
    # May be None for non-video or missing deps
    assert video_sample is not None



"""
Code adapated from https://github.com/mlfoundations/open_clip/blob/main/src/training/zero_shot.py
Thanks to the authors of OpenCLIP
"""

import logging
from contextlib import suppress

import torch
import torch.nn.functional as F
from sklearn.metrics import balanced_accuracy_score, classification_report
from tqdm import tqdm

# from open_clip import image_to_device


def zero_shot_classifier(model, tokenizer, classnames, templates, device, amp=True):
    """
    This function returns zero-shot vectors for each class in order
    to use it for zero-shot classification.


    model:
        CLIP-like model with `encode_text`

    tokenizer:
        text tokenizer, i.e. convert list of strings to torch.Tensor of integers

    classnames: list of str
        name of classes

    templates: list of str
        templates to use.

    Returns
    -------

    torch.Tensor of shape (N,C) where N is the number
    of templates, and C is the number of classes.
    """
    import os

    if os.environ.get("CLIP_BENCHMARK_DEBUG", None) == "1":
        breakpoint()

    autocast = torch.cuda.amp.autocast if amp else suppress
    with torch.no_grad(), autocast():
        zeroshot_weights = []
        for classname in tqdm(classnames):
            if type(templates) == dict:
                # class-specific prompts (e.g., CuPL https://arxiv.org/abs/2209.03320)
                texts = templates[classname]
            elif type(templates) == list:
                # generic prompts tht are specialized for each class by replacing {c} with the class name
                texts = [template.format(c=classname) for template in templates]
            else:
                raise ValueError("templates must be a list or a dict")
            texts = tokenizer(texts).to(device)  # tokenize
            class_embeddings = model.encode_text(texts)
            class_embedding = F.normalize(class_embeddings, dim=-1).mean(dim=0)
            class_embedding /= class_embedding.norm()
            zeroshot_weights.append(class_embedding)
        zeroshot_weights = torch.stack(zeroshot_weights, dim=1).to(device)
    return zeroshot_weights


def accuracy(output, target, topk=(1,)):
    """
    Compute top-k accuracy

    output: torch.Tensor
        shape (N, C) where N is the number of examples, C the number of classes.
        these are the logits.

    target: torch.Tensor
        shape (N,) where N is the number of examples. Groundtruth class id of each example.

    topk: tuple
        which topk to compute, e.g., topk=(1,5) will compute top-1 and top-5 accuracies

    Returns
    -------

    list of top-k accuracies in the same order as `topk`
    """
    pred = output.topk(max(topk), 1, True, True)[1].t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    n = len(target)
    return [
        float(correct[:k].reshape(-1).float().sum(0, keepdim=True).cpu().numpy()) / n
        for k in topk
    ]


def run_classification(
    model, classifier, dataloader, device, video_dataset=False, amp=True, args=None
):
    """
    Run zero-shot classifcation

    model: torch.nn.Module
        CLIP-like model with `encode_image` and `encode_text`

    classifier: torch.Tensor
        obtained from the function `zero_shot_classifier`

    dataloader: torch.utils.data.Dataloader

    Returns
    -------
    (pred, true)  where
        - pred (N, C) are the logits
        - true (N,) are the actual classes
    """
    autocast = torch.cuda.amp.autocast if amp else suppress
    pred = []
    true = []
    nb = 0
    with torch.no_grad():
        import os

        if os.environ.get("CLIP_BENCHMARK_DEBUG", None) == "1":
            breakpoint()

        for images, target in tqdm(dataloader):

            if isinstance(images, torch.Tensor):
                images = images.to(device, torch.float32)
            elif isinstance(images, (list, tuple)):  # video frames as a list/tuple
                images = [x.to(device, torch.float32) for x in images]
                images = torch.stack(images, dim=0).permute(1,0,2,3,4).contiguous() # nbchw -> bncwh
            else:
                raise NotImplementedError

            target = target.to(device)

            with autocast():
                # predict
                if video_dataset:
                    image_features = model.encode_video(images)
                else:
                    image_features = model.encode_image(images)

                image_features = F.normalize(image_features, dim=-1)
                logits = 100.0 * image_features @ classifier

            true.append(target.cpu())
            pred.append(logits.float().cpu())

    pred = torch.cat(pred)
    true = torch.cat(true)
    return pred, true


def average_precision_per_class(scores, targets):
    """
    Compute average precision  for each class
    this metric is used for multi-label classification
    see explanations here https://fangdahan.medium.com/calculate-mean-average-precision-map-for-multi-label-classification-b082679d31be
    Code is adapted from https://github.com/pytorch/tnt/blob/master/torchnet/meter/meter.py, thanks to the authors of `tnt`.

    Parameters
    ----------

    scores: torch.Tensor
        logits, of shape (N,C) where N is the number of examples, C the number of classes

    targets: torch.Tensor
        one-hot vectors of groundtruth targets (N, C), where N is the number of examples, C is the
        number of classes

    Returns
    -------

    torch.Tensor of shape (C,) of avereage precision for each class, where C is
    the number of classes.

    """
    ap = torch.zeros(scores.size(1))
    rg = torch.arange(1, scores.size(0) + 1).float()
    # compute average precision for each class
    for k in range(scores.size(1)):
        # sort scores
        scores_k = scores[:, k]
        targets_k = targets[:, k]
        _, sortind = torch.sort(scores_k, 0, True)
        truth = targets_k[sortind]
        tp = truth.float().cumsum(0)
        # compute precision curve
        precision = tp.div(rg)
        # compute average precision
        ap[k] = precision[truth.bool()].sum() / max(float(truth.sum()), 1)
    return ap


def evaluate(
    model,
    dataloader,
    tokenizer,
    classnames,
    templates,
    device,
    video_dataset=False,
    amp=True,
    verbose=False,
    save_clf=None,
    load_clfs=[],
    args=None,
):
    """
    Run zero-shot classification and evaluate the metrics

    Parameters
    ----------

    model: torch.nn.Module
        CLIP-like model with `encode_image` and `encode_text`

    dataloader: torch.utils.data.Dataloader

    tokenizer: text tokenizer

    classnames: list of str
        class names

    templates: list of str
        templates to use for zero-shot classification

    device: cpu/cuda

    amp: whether to use automatic mixed precision

    verbose: whether to use verbose model

    Returns
    -------

    dict of classification metrics
    """
    if len(load_clfs) > 0:
        n = len(load_clfs)
        classifier = torch.load(load_clfs[0], map_location="cpu") / n
        for i in range(1, n):
            classifier = classifier + torch.load(load_clfs[i], map_location="cpu") / n
        classifier = classifier.to(device)
    else:
        classifier = zero_shot_classifier(
            model, tokenizer, classnames, templates, device, amp=amp
        )

    if save_clf is not None:
        torch.save(classifier, save_clf)
        # exit() - not sure if we want to exit here or not.

    logits, target = run_classification(
        model,
        classifier,
        dataloader,
        device,
        video_dataset=video_dataset,
        amp=amp,
        args=args,
    )
    is_multilabel = len(target.shape) == 2

    if is_multilabel:
        if verbose:
            print("Detected a multi-label classification dataset")
        # Multiple labels per image, multiple classes on the dataset
        ap_per_class = average_precision_per_class(logits, target)
        if verbose:
            for class_name, ap in zip(
                dataloader.dataset.classes, ap_per_class.tolist()
            ):
                print(f"Class: {class_name}, AveragePrecision: {ap}")
        return {"mean_average_precision": ap_per_class.mean().item()}
    else:
        # Single label per image, multiple classes on the dataset
        # just compute accuracy and mean_per_class_recall

        pred = logits.argmax(axis=1)
        # measure accuracy
        if len(dataloader.dataset.classes) >= 5:
            acc1, acc5 = accuracy(logits, target, topk=(1, 5))
        else:
            (acc1,) = accuracy(logits, target, topk=(1,))
            acc5 = float("nan")
        mean_per_class_recall = balanced_accuracy_score(target, pred)
        if verbose:
            print(classification_report(target, pred, digits=3))
        return {
            "acc1": acc1,
            "acc5": acc5,
            "mean_per_class_recall": mean_per_class_recall,
        }



import copy
import re, ast
from transformers import AutoConfig, LlamaConfig
from transformers.configuration_utils import PretrainedConfig
from transformers.utils import logging

from easydict import EasyDict as MyEasyDict
from importlib import import_module
import os.path as osp
import argparse
import json
from copy import deepcopy
import sys


class VideoMAEv2Config(PretrainedConfig):
    model_type = 'VideoMAEv2_Base'
    def __init__(
            self,
            **kwargs):
        super().__init__(**kwargs)



"""UMTScore — Unified Multi-modal Transformer Score.

Video-text alignment scoring using UMT (Unified Multi-modal
Transformer) features. Measures how well video content matches
text descriptions via cross-modal similarity.

umtscore — higher = better alignment (0-1 range)
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class UMTScoreModule(PipelineModule):
    name = "umtscore"
    description = "UMTScore video-text alignment via UMT features"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._clip_model = None
        self._clip_processor = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try UMT model
        try:
            import umt
            self._model = umt
            self._backend = "native"
            logger.info("UMTScore (native) initialised")
            return
        except ImportError:
            pass

        # Tier 2: Try CLIP as proxy for cross-modal similarity
        try:
            from transformers import CLIPModel, CLIPProcessor
            self._clip_model = CLIPModel.from_pretrained(
                "openai/clip-vit-base-patch32"
            )
            self._clip_processor = CLIPProcessor.from_pretrained(
                "openai/clip-vit-base-patch32"
            )
            self._backend = "clip"
            logger.info("UMTScore (CLIP proxy) initialised")
            return
        except (ImportError, Exception):
            pass

        # Tier 3: Heuristic fallback
        self._backend = "heuristic"
        logger.info("UMTScore (heuristic) initialised — install transformers for CLIP proxy")

    def process(self, sample: Sample) -> Sample:
        try:
            caption = getattr(sample, "caption", None)
            if not caption:
                return sample

            caption_text = caption.text if hasattr(caption, "text") else str(caption)

            if self._backend == "native":
                score = float(self._model.score(str(sample.path), caption_text))
            elif self._backend == "clip":
                score = self._process_clip(sample, caption_text)
            else:
                score = self._process_heuristic(sample, caption_text)

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.umtscore = score

        except Exception as e:
            logger.warning(f"UMTScore failed for {sample.path}: {e}")

        return sample

    def _process_clip(self, sample: Sample, caption: str) -> Optional[float]:
        """CLIP-based text-video similarity proxy."""
        import torch
        from PIL import Image

        frames = self._extract_frames(sample)
        if not frames:
            return None

        scores = []
        for frame in frames:
            frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            pil_img = Image.fromarray(frame_rgb)

            inputs = self._clip_processor(
                text=[caption], images=pil_img, return_tensors="pt", padding=True
            )
            with torch.no_grad():
                outputs = self._clip_model(**inputs)

            # Cosine similarity from CLIP
            logits = outputs.logits_per_image
            sim = float(logits.softmax(dim=-1).max().item())
            scores.append(sim)

        return float(np.clip(np.mean(scores), 0.0, 1.0))

    def _process_heuristic(self, sample: Sample, caption: str) -> Optional[float]:
        """Heuristic: simple content-text alignment estimation."""
        frames = self._extract_frames(sample)
        if not frames:
            return None

        # Analyze visual content features
        visual_features = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            b, g, r = (
                frame[:, :, 0].astype(np.float64),
                frame[:, :, 1].astype(np.float64),
                frame[:, :, 2].astype(np.float64),
            )

            # Basic visual descriptors
            brightness = np.mean(gray) / 255.0
            contrast = np.std(gray) / 128.0
            colorfulness = (np.std(r - g) + np.std(0.5 * (r + g) - b)) / 256.0
            sharpness = min(cv2.Laplacian(gray, cv2.CV_64F).var() / 1000.0, 1.0)
            visual_features.append([brightness, contrast, colorfulness, sharpness])

        avg_features = np.mean(visual_features, axis=0)

        # Text analysis heuristic: longer, more descriptive captions suggest
        # more alignment potential; visual richness correlates with alignment
        words = caption.lower().split()
        text_richness = min(len(words) / 20.0, 1.0)
        visual_richness = float(np.mean(avg_features))

        # Heuristic alignment: rich visuals + rich text = likely aligned
        score = 0.5 * visual_richness + 0.3 * text_richness + 0.2

        return float(np.clip(score, 0.0, 1.0))

    def _extract_frames(self, sample: Sample) -> list:
        """Extract frames from video or image."""
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return frames
                indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        frames.append(frame)
            finally:
                cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)

        return frames



"""Tests for face_cross_similarity module."""

import numpy as np
import pytest

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics, Sample, DatasetStats


def test_face_cross_similarity_basics():
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    _test_module_basics(FaceCrossSimilarityModule, "face_cross_similarity")


def test_face_cross_similarity_image(image_sample):
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    image_sample.quality_metrics = QualityMetrics()
    m = FaceCrossSimilarityModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample


def test_face_cross_similarity_video(video_sample):
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    video_sample.quality_metrics = QualityMetrics()
    m = FaceCrossSimilarityModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample


def test_face_cross_similarity_empty_cache():
    """post_process with empty cache should not crash."""
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    m = FaceCrossSimilarityModule()
    # Empty cache, should skip gracefully
    m.post_process([])


def test_face_cross_similarity_cosine_math():
    """Verify cosine similarity computation using synthetic embeddings."""
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    m = FaceCrossSimilarityModule()

    # Two identical embeddings should have similarity = 1.0
    emb = np.array([1.0, 0.0, 0.0], dtype=np.float32)
    m._embeddings_cache["a"] = [emb]
    m._embeddings_cache["b"] = [emb]

    sample_a = Sample(path="a", is_video=False)
    sample_a.quality_metrics = QualityMetrics()
    sample_b = Sample(path="b", is_video=False)
    sample_b.quality_metrics = QualityMetrics()

    m.post_process([sample_a, sample_b])

    assert sample_a.quality_metrics.face_cross_similarity is not None
    assert abs(sample_a.quality_metrics.face_cross_similarity - 1.0) < 1e-5
    assert abs(sample_b.quality_metrics.face_cross_similarity - 1.0) < 1e-5


def test_face_cross_similarity_orthogonal():
    """Orthogonal embeddings should have similarity close to 0."""
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    m = FaceCrossSimilarityModule()

    emb_a = np.array([1.0, 0.0, 0.0], dtype=np.float32)
    emb_b = np.array([0.0, 1.0, 0.0], dtype=np.float32)
    m._embeddings_cache["a"] = [emb_a]
    m._embeddings_cache["b"] = [emb_b]

    sample_a = Sample(path="a", is_video=False)
    sample_a.quality_metrics = QualityMetrics()
    sample_b = Sample(path="b", is_video=False)
    sample_b.quality_metrics = QualityMetrics()

    m.post_process([sample_a, sample_b])

    assert sample_a.quality_metrics.face_cross_similarity is not None
    assert abs(sample_a.quality_metrics.face_cross_similarity) < 1e-5


def test_face_cross_similarity_clustering():
    """Test identity clustering with synthetic embeddings."""
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    m = FaceCrossSimilarityModule({"similarity_threshold": 0.9})

    # Two clusters: (a, b) similar, (c) different
    emb1 = np.array([1.0, 0.0, 0.0], dtype=np.float32)
    emb2 = np.array([0.99, 0.1, 0.0], dtype=np.float32)
    emb2 = emb2 / np.linalg.norm(emb2)
    emb3 = np.array([0.0, 0.0, 1.0], dtype=np.float32)

    m._embeddings_cache["a"] = [emb1]
    m._embeddings_cache["b"] = [emb2]
    m._embeddings_cache["c"] = [emb3]

    sample_a = Sample(path="a", is_video=False)
    sample_a.quality_metrics = QualityMetrics()
    sample_b = Sample(path="b", is_video=False)
    sample_b.quality_metrics = QualityMetrics()
    sample_c = Sample(path="c", is_video=False)
    sample_c.quality_metrics = QualityMetrics()

    # Verify clustering logic directly
    # emb1 . emb2 = 0.99 / sqrt(0.99^2 + 0.1^2) ~ 0.995 -> same cluster
    # emb1 . emb3 = 0 -> different cluster
    sim_matrix = np.zeros((3, 3))
    embs = [emb1, emb2, emb3]
    for i in range(3):
        for j in range(3):
            sim_matrix[i, j] = float(np.dot(embs[i], embs[j]))

    clusters = m._cluster_identities(sim_matrix)
    assert clusters == 2, f"Expected 2 clusters, got {clusters}"


def test_face_cross_similarity_fields():
    """Verify fields exist in QualityMetrics and DatasetStats."""
    # QualityMetrics fields
    qm = QualityMetrics()
    assert hasattr(qm, "face_cross_similarity")
    assert hasattr(qm, "face_identity_count")
    assert qm.face_cross_similarity is None
    assert qm.face_identity_count is None

    # DatasetStats fields
    ds = DatasetStats(total_samples=0, valid_samples=0, invalid_samples=0, total_size=0)
    assert hasattr(ds, "face_similarity_matrix")
    assert hasattr(ds, "avg_face_cross_similarity")
    assert hasattr(ds, "identity_cluster_count")


def test_face_cross_similarity_field_groups():
    """Verify field group assignments."""
    qm = QualityMetrics()
    groups = qm._FIELD_GROUPS
    assert groups.get("face_cross_similarity") == "face"
    assert groups.get("face_identity_count") == "face"


def test_face_cross_similarity_max_cache():
    """Module should stop caching when max_cache_size is reached."""
    from ayase.modules.face_cross_similarity import FaceCrossSimilarityModule

    m = FaceCrossSimilarityModule({"max_cache_size": 2})
    m._backend = "insightface"  # pretend we have a backend
    # Fill cache to max
    m._embeddings_cache["a"] = [np.zeros(3)]
    m._embeddings_cache["b"] = [np.zeros(3)]

    # Process should skip because cache is full
    sample = Sample(path="c", is_video=False)
    sample.quality_metrics = QualityMetrics()
    result = m.process(sample)
    assert "c" not in m._embeddings_cache



"""Tests for butteraugli module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_butteraugli_basics():
    from ayase.modules.butteraugli import ButteraugliModule
    _test_module_basics(ButteraugliModule, "butteraugli")

def test_butteraugli_no_reference(image_sample):
    from ayase.modules.butteraugli import ButteraugliModule
    m = ButteraugliModule()
    result = m.process(image_sample)
    assert result is image_sample



"""Tests for video_text_matching module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_video_text_matching_basics():
    from ayase.modules.video_text_matching import VideoTextMatchingModule
    _test_module_basics(VideoTextMatchingModule, "video_text_matching")

def test_video_text_matching_image(image_sample):
    from ayase.modules.video_text_matching import VideoTextMatchingModule
    image_sample.quality_metrics = QualityMetrics()
    m = VideoTextMatchingModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_video_text_matching_video(video_sample):
    from ayase.modules.video_text_matching import VideoTextMatchingModule
    video_sample.quality_metrics = QualityMetrics()
    m = VideoTextMatchingModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""GAMIVAL — Gaming Video Quality Assessment.

Yu et al. 2023 — NR-VQA specifically designed for cloud gaming content.
Combines NSS features with CNN features tailored for gaming artifacts
like encoding distortions, frame drops, and rendering issues.

gamival_score — higher = better quality
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


def _gaming_artifact_features(frame: np.ndarray) -> np.ndarray:
    """Extract features relevant to gaming content quality."""
    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
    h, w = gray.shape

    # Sharpness (gaming content should be sharp)
    lap_var = cv2.Laplacian(gray, cv2.CV_64F).var()

    # Blockiness detection (common in compressed gaming streams)
    block_size = 8
    h_blocks = h // block_size
    w_blocks = w // block_size
    blockiness = 0.0
    if h_blocks > 1 and w_blocks > 1:
        cropped = gray[: h_blocks * block_size, : w_blocks * block_size]
        # Measure discontinuities across block boundaries
        h_edges = np.abs(cropped[block_size-1::block_size, :].astype(float) - cropped[block_size::block_size, :].astype(float))
        v_edges = np.abs(cropped[:, block_size-1::block_size].astype(float) - cropped[:, block_size::block_size].astype(float))
        blockiness = float(np.mean(h_edges) + np.mean(v_edges))

    # Banding detection (gradient posterization common in gaming)
    gradient_y = np.abs(np.diff(gray, axis=0))
    gradient_x = np.abs(np.diff(gray, axis=1))
    # Low gradient areas that should be smooth
    smooth_mask_y = gradient_y < 2.0
    smooth_mask_x = gradient_x < 2.0
    banding = float(np.mean(smooth_mask_y) + np.mean(smooth_mask_x)) / 2.0

    # Colorfulness (gaming often has vivid colors)
    b, g, r = (
        frame[:, :, 0].astype(np.float64),
        frame[:, :, 1].astype(np.float64),
        frame[:, :, 2].astype(np.float64),
    )
    rg = r - g
    yb = 0.5 * (r + g) - b
    colorfulness = np.sqrt(rg.var() + yb.var()) + 0.3 * np.sqrt(
        rg.mean() ** 2 + yb.mean() ** 2
    )

    # Contrast
    contrast = gray.std()

    # NSS features (MSCN)
    mu = cv2.GaussianBlur(gray, (7, 7), 7 / 6)
    sigma = np.sqrt(
        np.abs(cv2.GaussianBlur(gray * gray, (7, 7), 7 / 6) - mu * mu)
    )
    sigma = np.maximum(sigma, 1e-7)
    mscn = (gray - mu) / sigma
    nss_mean = float(np.mean(np.abs(mscn)))
    nss_var = float(np.var(mscn))

    return np.array([
        lap_var,        # sharpness
        blockiness,     # block artifacts
        banding,        # gradient banding
        colorfulness,   # color saturation
        contrast,       # luminance contrast
        nss_mean,       # NSS regularity
        nss_var,        # NSS variance
    ], dtype=np.float64)


class GAMIVALModule(PipelineModule):
    name = "gamival"
    description = "GAMIVAL cloud gaming NR-VQA with NSS + CNN features (2023)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try native GAMIVAL
        try:
            import gamival
            self._model = gamival
            self._backend = "native"
            logger.info("GAMIVAL (native) initialised")
            return
        except ImportError:
            pass

        # Tier 2: Heuristic fallback
        self._backend = "heuristic"
        logger.info("GAMIVAL (heuristic) initialised — install gamival for full model")

    def process(self, sample: Sample) -> Sample:
        try:
            if self._backend == "native":
                score = float(self._model.predict(str(sample.path)))
            else:
                score = self._process_heuristic(sample)

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.gamival_score = score

        except Exception as e:
            logger.warning(f"GAMIVAL failed for {sample.path}: {e}")

        return sample

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: gaming-specific artifact detection + NSS features."""
        frames = []

        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return None
                indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        frames.append(frame)
            finally:
                cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)

        if not frames:
            return None

        # Extract gaming-specific features per frame
        all_features = []
        for frame in frames:
            all_features.append(_gaming_artifact_features(frame))

        feat_matrix = np.array(all_features)
        feat_mean = np.mean(feat_matrix, axis=0)

        # Quality components
        sharpness = min(feat_mean[0] / 800.0, 1.0)
        # Low blockiness is good
        block_quality = 1.0 / (1.0 + feat_mean[1] * 0.05)
        # Low banding is good (banding ~0.5 means half pixels have near-zero gradient)
        band_quality = 1.0 - min(feat_mean[2], 1.0) * 0.5
        colorfulness = min(feat_mean[3] / 100.0, 1.0)
        contrast = min(feat_mean[4] / 65.0, 1.0)
        nss_regularity = 1.0 / (1.0 + abs(feat_mean[5] - 0.8))

        spatial_score = (
            0.25 * sharpness
            + 0.20 * block_quality
            + 0.10 * band_quality
            + 0.15 * colorfulness
            + 0.15 * contrast
            + 0.15 * nss_regularity
        )

        # Temporal quality (for video)
        if len(frames) > 1:
            # Frame drop detection: large sudden changes
            grays = [
                cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(np.float64) for f in frames
            ]
            diffs = []
            for i in range(len(grays) - 1):
                diffs.append(np.mean(np.abs(grays[i + 1] - grays[i])))
            diffs = np.array(diffs)
            # Consistency of motion (frame drops cause spikes)
            temporal = 1.0 / (1.0 + np.std(diffs) * 0.1)
            score = 0.75 * spatial_score + 0.25 * temporal
        else:
            score = spatial_score

        return float(np.clip(score, 0.0, 1.0))



import torch
import torch.nn as nn

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
backwarp_tenGrid = {}


def warp(tenInput, tenFlow):
    k = (str(tenFlow.device), str(tenFlow.size()))
    if k not in backwarp_tenGrid:
        tenHorizontal = torch.linspace(-1.0, 1.0, tenFlow.shape[3], device=device).view(
            1, 1, 1, tenFlow.shape[3]).expand(tenFlow.shape[0], -1, tenFlow.shape[2], -1)
        tenVertical = torch.linspace(-1.0, 1.0, tenFlow.shape[2], device=device).view(
            1, 1, tenFlow.shape[2], 1).expand(tenFlow.shape[0], -1, -1, tenFlow.shape[3])
        backwarp_tenGrid[k] = torch.cat(
            [tenHorizontal, tenVertical], 1).to(device)

    tenFlow = torch.cat([tenFlow[:, 0:1, :, :] / ((tenInput.shape[3] - 1.0) / 2.0),
                         tenFlow[:, 1:2, :, :] / ((tenInput.shape[2] - 1.0) / 2.0)], 1)

    g = (backwarp_tenGrid[k] + tenFlow).permute(0, 2, 3, 1)
    return torch.nn.functional.grid_sample(input=tenInput, grid=g, mode='bilinear', padding_mode='border', align_corners=True)



# Copyright (2024) Bytedance Ltd. and/or its affiliates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
class Color:

    @staticmethod
    def red(x):
        return '\33[31m' +x + '\033[0m'
    
    @staticmethod
    def green(x):
        return '\33[32m' +x + '\033[0m'

    @staticmethod
    def yellow(x):
        return '\33[33m' +x + '\033[0m'

    @staticmethod
    def blue(x):
        return '\33[34m' +x + '\033[0m'

    @staticmethod
    def violet(x):
        return '\33[35m' +x + '\033[0m'

    



"""Tests for lip_sync module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_lip_sync_basics():
    from ayase.modules.lip_sync import LipSyncModule
    _test_module_basics(LipSyncModule, "lip_sync")

def test_lip_sync_video(video_sample):
    from ayase.modules.lip_sync import LipSyncModule
    video_sample.quality_metrics = QualityMetrics()
    m = LipSyncModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Tests for sfid module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_sfid_basics():
    from ayase.modules.sfid import SFIDModule
    _test_module_basics(SFIDModule, "sfid")

def test_sfid_extract(video_sample):
    from ayase.modules.sfid import SFIDModule
    m = SFIDModule()
    feat = m.extract_features(video_sample)
    # May be None for non-video or missing deps
    assert video_sample is not None



"""SimpleVQA — Simple Blind Video Quality Assessment.

2022 — Swin Transformer-B spatial + fixed SlowFast temporal
features for blind VQA. Base model for RQ-VQA.

GitHub: https://github.com/sunwei925/SimpleVQA

simplevqa_score — higher = better quality
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class SimpleVQAModule(PipelineModule):
    name = "simplevqa"
    description = "SimpleVQA Swin+SlowFast blind VQA (2022)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._backend = "heuristic"

    def setup(self) -> None:
        try:
            import simplevqa
            self._model = simplevqa
            self._backend = "native"
            logger.info("SimpleVQA (native) initialised")
            return
        except ImportError:
            pass

        self._backend = "heuristic"
        logger.info("SimpleVQA (heuristic) — install simplevqa for full model")

    def process(self, sample: Sample) -> Sample:
        try:
            score = (
                float(self._model.predict(str(sample.path)))
                if self._backend == "native"
                else self._process_heuristic(sample)
            )

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.simplevqa_score = score

        except Exception as e:
            logger.warning(f"SimpleVQA failed for {sample.path}: {e}")

        return sample

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: spatial (Swin-like multi-scale) + temporal (SlowFast-like)."""
        frames = self._extract_frames(sample)
        if not frames:
            return None

        spatial_scores = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)

            # Multi-scale spatial (Swin-like pyramid)
            scale_scores = []
            for scale in [1.0, 0.5, 0.25]:
                h, w = gray.shape
                sh, sw = int(h * scale), int(w * scale)
                if sh < 16 or sw < 16:
                    continue
                scaled = cv2.resize(gray, (sw, sh))
                lap = cv2.Laplacian(scaled, cv2.CV_64F).var()
                scale_scores.append(min(lap / 500.0, 1.0))

            spatial_scores.append(np.mean(scale_scores) if scale_scores else 0.5)

        spatial = float(np.mean(spatial_scores))

        # SlowFast-like temporal: slow pathway (all frames) + fast pathway (frame diffs)
        if len(frames) > 1:
            # Slow: overall frame consistency
            slow_diffs = []
            for i in range(len(frames) - 1):
                g1 = cv2.cvtColor(frames[i], cv2.COLOR_BGR2GRAY).astype(float)
                g2 = cv2.cvtColor(frames[i + 1], cv2.COLOR_BGR2GRAY).astype(float)
                g1 = cv2.resize(g1, (160, 120))
                g2 = cv2.resize(g2, (160, 120))
                slow_diffs.append(np.mean(np.abs(g1 - g2)))

            slow_quality = 1.0 / (1.0 + np.var(slow_diffs) * 0.01)

            # Fast: rapid motion changes
            if len(slow_diffs) > 1:
                diff_of_diffs = np.diff(slow_diffs)
                fast_quality = 1.0 / (1.0 + np.var(diff_of_diffs) * 0.05)
            else:
                fast_quality = 1.0

            temporal = 0.6 * slow_quality + 0.4 * fast_quality
        else:
            temporal = 1.0

        score = 0.65 * spatial + 0.35 * temporal
        return float(np.clip(score, 0.0, 1.0))

    def _extract_frames(self, sample: Sample):
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return []
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)
        return frames



"""World Consistency Score — object permanence + causal compliance (2025).

Paper: https://arxiv.org/abs/2508.00144
world_consistency_score — higher = better
"""
import logging, cv2, numpy as np
from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule
logger = logging.getLogger(__name__)

class WorldConsistencyModule(PipelineModule):
    name = "world_consistency"
    description = "World Consistency Score: object permanence + causal compliance (2025)"
    default_config = {"subsample": 12}
    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 12)
        self._backend = "heuristic"
    def setup(self):
        try:
            import wcs; self._model = wcs; self._backend = "native"; return
        except ImportError: pass
        self._backend = "heuristic"
    def process(self, sample):
        if not sample.is_video: return sample
        try:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 1: cap.release(); return sample
            indices = np.linspace(0, total-1, min(self.subsample, total), dtype=int)
            frames = []
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx); ret, f = cap.read()
                if ret: frames.append(cv2.resize(f, (320, 240)))
            cap.release()
            if len(frames) < 2: return sample

            # Object permanence: structural similarity across frames
            ssim_scores = []
            for i in range(len(frames)-1):
                g1 = cv2.cvtColor(frames[i], cv2.COLOR_BGR2GRAY).astype(float)
                g2 = cv2.cvtColor(frames[i+1], cv2.COLOR_BGR2GRAY).astype(float)
                mu1, mu2 = g1.mean(), g2.mean()
                s1, s2 = g1.var(), g2.var()
                cov = np.mean((g1-mu1)*(g2-mu2))
                c1, c2 = 6.5025, 58.5225
                ssim = ((2*mu1*mu2+c1)*(2*cov+c2))/((mu1**2+mu2**2+c1)*(s1+s2+c2))
                ssim_scores.append(max(ssim, 0))
            permanence = float(np.mean(ssim_scores))

            # Relation stability: edge structure consistency
            edge_counts = [np.mean(cv2.Canny(f, 50, 150) > 0) for f in frames]
            stability = 1.0/(1.0+np.var(edge_counts)*100)

            # Flicker penalty
            diffs = [np.mean(np.abs(cv2.cvtColor(frames[i], cv2.COLOR_BGR2GRAY).astype(float) - cv2.cvtColor(frames[i+1], cv2.COLOR_BGR2GRAY).astype(float))) for i in range(len(frames)-1)]
            flicker = 1.0/(1.0+np.var(diffs)*0.01)

            score = 0.40*permanence + 0.30*stability + 0.30*flicker
            if sample.quality_metrics is None: sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.world_consistency_score = float(np.clip(score, 0, 1))
        except Exception as e:
            logger.warning(f"WorldConsistency failed: {e}")
        return sample



"""Tests for beat_alignment module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_beat_alignment_basics():
    from ayase.modules.beat_alignment import BeatAlignmentModule
    _test_module_basics(BeatAlignmentModule, "beat_alignment")

def test_beat_alignment_video(video_sample):
    from ayase.modules.beat_alignment import BeatAlignmentModule
    video_sample.quality_metrics = QualityMetrics()
    m = BeatAlignmentModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



# Copyright (c) Meta Platforms, Inc. and affiliates.

import itertools
import json
import logging
import os
import re
import traceback
from typing import Any, Callable, Dict, Iterator, List, Optional, cast

import numpy as np
import torch
from torch.utils.data import IterableDataset, get_worker_info

logger = logging.getLogger(__name__)


def get_worker_info():
    worker_info = torch.utils.data.get_worker_info()
    if worker_info is None:
        num_workers = 1
        worker_id = 0
    else:
        num_workers = worker_info.num_workers
        worker_id = worker_info.id

    return worker_id, num_workers


def get_global_rank_info(rank, world_size):
    worker_id, num_workers = get_worker_info()
    dataloader_rank = rank * num_workers + worker_id
    dataloader_world_size = world_size * num_workers
    return dataloader_rank, dataloader_world_size


class JSONLIterator:
    def __init__(
        self,
        fpath: str,
        world_size: int,
        world_rank: int,
        infinite: bool,
    ):
        assert 0 <= world_rank < world_size, (world_rank, world_size)
        self.f = open(fpath, "r", encoding="utf-8")
        self.fpath = fpath
        self.world_size = world_size
        self.world_rank = world_rank
        self.line_num = 0
        self.iter = iter(self.gen(infinite))
        self.iter_id = 0

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iter)

    def gen(self, infinite: bool) -> Iterator[Dict]:
        while True:
            if self.world_rank == 0:
                logger.info(f"Starting iteration {self.iter_id} over {self.fpath} ...")
            self.iter_id += 1
            while True:
                line, self.line_num = self.f.readline(), self.line_num + 1
                if not line:
                    break
                if (self.line_num - 1) % self.world_size == self.world_rank:
                    yield json.loads(line)
            if not infinite:
                break
            self.set_position(None)
        self.f.close()

    def set_position(self, position: Optional[int]):
        logger.warning(
            f"Setting JSONL position on {self.fpath} "
            f"({self.world_rank}/{self.world_size}): {position}"
        )
        if position is None:
            self.f.seek(0)
            self.line_num = 0
        else:
            assert isinstance(position, int)
            self.f.seek(position)
            self.line_num = (
                self.world_rank + 1
            )  # Restore value of line_num (modulo world_size)

    def get_position(self) -> Optional[int]:
        file_pos = self.f.tell()
        if file_pos == 0 and self.line_num == 0:
            return None
        assert (self.line_num - 1) % self.world_size == self.world_rank
        return file_pos

    def get_example_file(self):
        """
        Return the path to a sample file to infer the content key
        """
        return self.fpath

    def get_id(self):
        """
        Return an identifier for the dataset this iterator represents
        """
        return self.fpath


class JSONLDirectoryIterator:
    """
    The JSONLDirectoryIterator is a data wrapper around a dataset folder, which contains
    multiple JSONL files. Internally, it reuses the JSONLIterator class to iterate through
    each individual file, and then wraps onto the next file once the current one is exhausted.

    Once all files in the directory have been iterated over, we wrap back to the first file
    ( if infinite is true ).

    This enables us to iterate over a dataset one chunk at a time.

    Also, note that we open the next chunk file on an ondemand basis, which means that we can
    modify chunks mid training as well to add more data, fix issues, etc.
    """

    def __init__(
        self,
        dirpath: str,
        world_size: int,
        world_rank: int,
        infinite: bool,
    ):
        assert 0 <= world_rank < world_size, (world_rank, world_size)
        self.dirpath = dirpath
        self.world_size = world_size
        self.world_rank = world_rank

        fnames = [
            x
            for x in os.listdir(self.dirpath)
            if re.fullmatch(r".*chunk\.\d+.*\.jsonl", x)
        ]
        self.fpaths = [os.path.join(self.dirpath, fname) for fname in sorted(fnames)]
        assert (
            len(self.fpaths) > 0
        ), f"Specified dataset location {self.dirpath} is empty."

        # Generator for cycling through the list of files
        if infinite:
            self.fpaths_generator = cast(Iterator[str], itertools.cycle(self.fpaths))
        else:
            self.fpaths_generator = cast(Iterator[str], iter(self.fpaths))

        self.iter = iter(self.gen(infinite))
        self.jsonl_iterator: Optional[JSONLIterator] = None

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iter)

    def gen(self, infinite: bool) -> Iterator[Dict]:
        # Handle the case when we're reloading from a saved state.
        if self.jsonl_iterator is not None:
            yield from self.jsonl_iterator

        for fpath in self.fpaths_generator:
            # Note that we set infinite to false here, because JSONLDirectoryIterator would take care of infinite looping
            self.jsonl_iterator = JSONLIterator(
                fpath,
                world_size=self.world_size,
                world_rank=self.world_rank,
                infinite=False,
            )

            yield from self.jsonl_iterator

    def set_position(self, state: Dict[str, Any]):
        logger.warning(
            f"Setting JSONL position on {self.dirpath} "
            f"({self.world_rank}/{self.world_size}): {state}"
        )
        fpath: Optional[str] = state["fpath"]
        position: Optional[int] = state["position"]
        if fpath is None or position is None:
            return

        assert isinstance(fpath, str)
        assert isinstance(position, int)

        # Fast forward the generator
        for fpath_candidate in self.fpaths_generator:
            if fpath_candidate == fpath:
                break

        # Create the JSONL iterator and set it's position appropriately
        self.jsonl_iterator = JSONLIterator(
            fpath,
            world_size=self.world_size,
            world_rank=self.world_rank,
            infinite=False,
        )
        self.jsonl_iterator.set_position(position)

    def get_position(self):
        if self.jsonl_iterator is None:
            return {
                "fpath": None,
                "position": None,
            }
        return {
            "fpath": self.jsonl_iterator.fpath,
            "position": self.jsonl_iterator.get_position(),
        }

    def get_example_file(self):
        """
        Return the path to a sample file to infer the content key
        """
        return self.fpaths[0]

    def get_id(self):
        """
        Return an identifier for the dataset this iterator represents
        """
        return self.dirpath


class IterativeJSONLDataset(IterableDataset):
    def __init__(
        self,
        global_rank: int,
        world_size: int,
        dataset_name: str,
        seed: int = 0,
        dataset_configs: Dict[str, Any] = {},
    ):
        self._dataset_name = dataset_name
        self._seed = seed
        self._dataset_conf = dataset_configs[dataset_name]

        self.global_rank = global_rank
        self.world_size = world_size
        self.data_path = self._dataset_conf.annotation

    def worker_init(self, worker_id, num_workers):
        dataloader_rank = self.global_rank * num_workers + worker_id
        dataloader_world_size = self.world_size * num_workers
        if os.path.isfile(self.data_path):
            self.jsonl_iterator = JSONLIterator(
                self.data_path,
                world_size=dataloader_world_size,
                world_rank=dataloader_rank,
                infinite=True,
            )
        else:
            self.jsonl_iterator = JSONLDirectoryIterator(
                dirpath=self.data_path,
                world_size=dataloader_world_size,
                world_rank=dataloader_rank,
                infinite=True,
            )
        if worker_id == 0:
            logger.info(
                f"Initializing JSONLDataset {self._dataset_name} on "
                f"dataloader rank {dataloader_rank} and world size {dataloader_world_size}"
            )

    def state_dict(self):
        pos = self.jsonl_iterator.get_position()
        if isinstance(pos, Dict):
            return pos
        else:
            return {"single_jsonl_position": pos}

    def load_state_dict(self, state_dict):
        if "single_jsonl_position" in state_dict:
            self.jsonl_iterator.set_position(state_dict["single_jsonl_position"])
        else:
            self.jsonl_iterator.set_position(state_dict)
        logger.info(f"JSONLDataset {self._dataset_name} resuming from {state_dict}.")

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.jsonl_iterator)


class DatasetMixer(IterableDataset):
    def __init__(
        self,
        mix: str,
        global_rank: int,
        world_size: int,
        seed: int = 0,
        preprocessors: List[Callable] = [],
        dataset_configs: Dict[str, Any] = {},
    ):
        super().__init__()

        self.dataset_and_preprocessors = []
        self.weights = []
        self.dataset_names = []
        self.totals = []

        self.global_rank = global_rank
        self.world_size = world_size
        self.seed = seed

        mix = "".join(mix.split())  # Remove whitespace

        for elem in mix.split(","):
            ds, weight = elem.split(":")

            if ds not in dataset_configs:
                raise ValueError(f"Dataset {ds} not found in dataset_configs.")
            if ds in self.dataset_names:
                raise ValueError(
                    f"Dataset {ds} already in the mix. Each dataset can only be used once."
                )

            dataset = IterativeJSONLDataset(
                global_rank=global_rank,
                world_size=world_size,
                dataset_name=ds,
                seed=seed,
                dataset_configs=dataset_configs,
            )
            _preprocessors = [
                p(dataset_config=dataset_configs[ds]) for p in preprocessors
            ]

            self.dataset_and_preprocessors.append((dataset, _preprocessors))
            self.weights.append(float(weight))
            self.dataset_names.append(ds)
            self.totals.append(0)

        self.weights = [w / sum(self.weights) for w in self.weights]
        self.rng = None

    def state_dict(self):
        return {
            "datasets": {
                ds_name: ds.state_dict()
                for ds_name, (ds, _) in zip(
                    self.dataset_names, self.dataset_and_preprocessors
                )
            },
            "totals": {
                ds_name: total
                for ds_name, total in zip(self.dataset_names, self.totals)
            },
            "rng": (
                [
                    s.tolist() if isinstance(s, np.ndarray) else s
                    for s in self.rng.get_state()
                ]
                if self.rng is not None
                else None
            ),
        }

    def load_state_dict(self, state_dict):
        for ds_name, sd in state_dict["datasets"].items():
            if ds_name in self.dataset_names:
                ds_idx = self.dataset_names.index(ds_name)
                ds, _ = self.dataset_and_preprocessors[ds_idx]
                ds.load_state_dict(sd)
                self.totals[ds_idx] = state_dict["totals"][ds_name]

        logger.info(
            f"DatasetMixer with datasets {self.dataset_names} resuming with total samples seen {self.totals} on process {os.getpid()}."
        )

        if state_dict["rng"] is not None:
            self.rng = np.random.RandomState()
            rng_state = [
                np.array(s) if isinstance(s, list) else s for s in state_dict["rng"]
            ]
            self.rng.set_state(rng_state)

    def worker_init(self, worker_id):
        worker_info = torch.utils.data.get_worker_info()
        for dataset, _ in self.dataset_and_preprocessors:
            if hasattr(dataset, "worker_init"):
                dataset.worker_init(worker_id, worker_info.num_workers)

    def __iter__(self):
        if self.rng is None:
            rank, world_size = get_global_rank_info(self.global_rank, self.world_size)
            self.rng = np.random.RandomState((rank, world_size, self.seed))

        while True:
            try:
                src_id = self.rng.choice(len(self.weights), p=self.weights)
                dataset, preprocessors = self.dataset_and_preprocessors[src_id]
                out = next(dataset)
                for preprocessor in preprocessors:
                    if out is not None:
                        out = preprocessor(out, self.rng)

                if out is None:
                    continue

                self.totals[src_id] += 1
                yield out
            except Exception as e:
                logger.error(
                    f"Error while iterating over dataset {self.dataset_names[src_id]}: {e}\n"
                    f"Traceback:\n{traceback.format_exc()}"
                )


class PersistentDataLoader:
    """
    A _very_ persistent dataloader.

    Uses StatefulDataLoader to save dataset state (make sure dataset has a state_dict() and load_state_dict() method).
    Also keeps the dataloader iterator and the epoch iterator separate, so that the dataloader workers are persistent.

    Also laughs in the face of torch when it tries to kill the whole job because a worker died. Instead, this dataloader
    will just gracefully restart the underlying iterator and correpsonding workers, while additionally loading the state dict
    so that it resumes from where it left off.

    This may or may not be a good idea.
    """

    def __init__(
        self,
        dataset,
        batch_size,
        workers,
        collate_fn=None,
        positions=None,
    ):
        from torchdata.stateful_dataloader import StatefulDataLoader

        self.dataloader = StatefulDataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=False,
            num_workers=workers,
            # pin_memory=True,
            multiprocessing_context="fork" if workers > 0 else None,
            collate_fn=collate_fn,
            worker_init_fn=(
                dataset.worker_init if hasattr(dataset, "worker_init") else None
            ),
            # persistent_workers=(workers > 0),
            snapshot_every_n_steps=1,
        )

        if positions is not None:
            self.load_state_dict(positions)

        self._dataloader_iter = iter(self.dataloader)

        # # Stop torch from killing us all
        # register_subscriber(self)

    def state_dict(self):
        return self.dataloader.state_dict()

    def load_state_dict(self, state_dict):
        self.dataloader.load_state_dict(state_dict)

    def __del__(self):
        pass  # unregister_subscriber(self)

    def __len__(self):
        return len(self.dataloader)

    def __iter__(self):
        self.iter = self.gen()
        return self

    def __next__(self):
        return next(self.iter)

    def _refresh_iter(self):
        # Called by the signal handler when a worker dies
        self._dataloader_iter = None

    def _get_next_sample(self):
        if self._dataloader_iter is None:
            self.dataloader.load_state_dict(self.dataloader.state_dict())
            self._dataloader_iter = iter(self.dataloader)

        try:
            return next(self._dataloader_iter)
        except (KeyboardInterrupt, StopIteration):
            raise
        except Exception as e:
            if self._dataloader_iter is None:
                # An interrupt forced us to respawn the dataloaders, do it next sample
                return self._get_next_sample()
            else:
                raise e

    def gen(self):
        while True:
            try:
                yield self._get_next_sample()
            except StopIteration:
                raise



import argparse
import os
from pathlib import Path

import torch
from omegaconf import OmegaConf

from ...apps.plm.transformer import LMTransformer, LMTransformerArgs
from ...core.args import dataclass_from_dict
from ...core.checkpoint import load_from_checkpoint


def build_model(
    ref_model_path: str,
    model_cls=LMTransformer,
    model_args_cls=LMTransformerArgs,
):
    ckpt_path = Path(ref_model_path)
    config = ckpt_path / "params.json"
    config = OmegaConf.load(config)

    model_args = dataclass_from_dict(model_args_cls, config.model, strict=False)
    model = model_cls(model_args)
    return model


def main():
    parser = argparse.ArgumentParser(description="Consolidate PLM checkpoints")
    parser.add_argument(
        "--ckpt",
        type=str,
        required=True,
        help="Path to the checkpoint directory to consolidate",
    )
    args = parser.parse_args()

    model = build_model(ref_model_path=args.ckpt)
    load_from_checkpoint(
        ckpt_dir=args.ckpt,
        model=model,
        optimizer=None,
        model_key="model",
    )

    consolidated_model_state_dict = model.state_dict()
    output_file = os.path.join(args.ckpt, "consolidated.pth")

    # Save the consolidated model state_dict using torch.save
    print(f"Saving consolidated model state_dict to: {output_file}")
    torch.save(consolidated_model_state_dict, output_file)
    print("Consolidated checkpoint saved successfully.")


if __name__ == "__main__":
    main()



from typing import List

from .score import Score

from .constants import HF_CACHE_DIR

from .models.itmscore_models import list_all_itmscore_models, get_itmscore_model

class ITMScore(Score):
    def prepare_scoremodel(self,
                           model='blip2-itm',
                           device='cuda',
                           cache_dir=HF_CACHE_DIR):
        return get_itmscore_model(
            model,
            device=device,
            cache_dir=cache_dir
        )
            
    def list_all_models(self) -> List[str]:
        return list_all_itmscore_models()